@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2014 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2015 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,7 +21,7 @@ public:
 	MODCONSTRUCTOR(CFailToBanMod) {}
 	virtual ~CFailToBanMod() {}
 
-	virtual bool OnLoad(const CString& sArgs, CString& sMessage) {
+	virtual bool OnLoad(const CString& sArgs, CString& sMessage) override {
 		CString sTimeout = sArgs.Token(0);
 		CString sAttempts = sArgs.Token(1);
 		unsigned int timeout = sTimeout.ToUInt();
@@ -46,7 +46,7 @@ public:
 		return true;
 	}
 
-	virtual void OnPostRehash() {
+	virtual void OnPostRehash() override {
 		m_Cache.Clear();
 	}
 
@@ -54,13 +54,13 @@ public:
 		m_Cache.AddItem(sHost, count, m_Cache.GetTTL());
 	}
 
-	virtual void OnModCommand(const CString& sCommand) {
+	virtual void OnModCommand(const CString& sCommand) override {
 		PutModule("This module can only be configured through its arguments.");
 		PutModule("The module argument is the number of minutes an IP");
 		PutModule("is blocked after a failed login.");
 	}
 
-	virtual void OnClientConnect(CZNCSock* pClient, const CString& sHost, unsigned short uPort) {
+	virtual void OnClientConnect(CZNCSock* pClient, const CString& sHost, unsigned short uPort) override {
 		unsigned int *pCount = m_Cache.GetItem(sHost);
 		if (sHost.empty() || pCount == NULL || *pCount < m_uiAllowedFailed) {
 			return;
@@ -73,7 +73,7 @@ public:
 		pClient->Close(Csock::CLT_AFTERWRITE);
 	}
 
-	virtual void OnFailedLogin(const CString& sUsername, const CString& sRemoteIP) {
+	virtual void OnFailedLogin(const CString& sUsername, const CString& sRemoteIP) override {
 		unsigned int *pCount = m_Cache.GetItem(sRemoteIP);
 		if (pCount)
 			Add(sRemoteIP, *pCount + 1);
@@ -81,7 +81,7 @@ public:
 			Add(sRemoteIP, 1);
 	}
 
-	virtual EModRet OnLoginAttempt(CSmartPtr<CAuthBase> Auth) {
+	virtual EModRet OnLoginAttempt(std::shared_ptr<CAuthBase> Auth) override {
 		// e.g. webadmin ends up here
 		const CString& sRemoteIP = Auth->GetRemoteIP();
 
@@ -109,4 +109,4 @@ template<> void TModInfo<CFailToBanMod>(CModInfo& Info) {
 	Info.SetArgsHelpText("You might enter the time in minutes for the IP banning and the number of failed logins before any action is taken.");
 }
 
-GLOBALMODULEDEFS(CFailToBanMod, "Block IPs for some time after a failed login")
+GLOBALMODULEDEFS(CFailToBanMod, "Block IPs for some time after a failed login.")
