@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2014 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2015 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -34,15 +34,20 @@ private:
 
 class CKeepNickMod : public CModule {
 public:
-	MODCONSTRUCTOR(CKeepNickMod) {}
+	MODCONSTRUCTOR(CKeepNickMod) {
+		AddHelpCommand();
+		AddCommand("Enable", static_cast<CModCommand::ModCmdFunc>(&CKeepNickMod::OnEnableCommand), "", "Try to get your primary nick");
+		AddCommand("Disable", static_cast<CModCommand::ModCmdFunc>(&CKeepNickMod::OnDisableCommand), "", "No longer trying to get your primary nick");
+		AddCommand("State", static_cast<CModCommand::ModCmdFunc>(&CKeepNickMod::OnStateCommand), "", "Show the current state");
+	}
 
 	~CKeepNickMod() {}
 
-	bool OnLoad(const CString& sArgs, CString& sMessage) {
+	bool OnLoad(const CString& sArgs, CString& sMessage) override {
 		m_pTimer = NULL;
 
 		// Check if we need to start the timer
-		if (m_pNetwork->IsIRCConnected())
+		if (GetNetwork()->IsIRCConnected())
 			OnIRCConnected();
 
 		return true;
@@ -53,7 +58,7 @@ public:
 			// No timer means we are turned off
 			return;
 
-		CIRCSock* pIRCSock = m_pNetwork->GetIRCSock();
+		CIRCSock* pIRCSock = GetNetwork()->GetIRCSock();
 
 		if (!pIRCSock)
 			return;
@@ -66,8 +71,8 @@ public:
 	}
 
 	CString GetNick() {
-		CString sConfNick = m_pNetwork->GetNick();
-		CIRCSock* pIRCSock = m_pNetwork->GetIRCSock();
+		CString sConfNick = GetNetwork()->GetNick();
+		CIRCSock* pIRCSock = GetNetwork()->GetIRCSock();
 
 		if (pIRCSock)
 			sConfNick = sConfNick.Left(pIRCSock->GetMaxNickLen());
@@ -75,8 +80,8 @@ public:
 		return sConfNick;
 	}
 
-	void OnNick(const CNick& Nick, const CString& sNewNick, const vector<CChan*>& vChans) {
-		if (sNewNick == m_pNetwork->GetIRCSock()->GetNick()) {
+	void OnNick(const CNick& Nick, const CString& sNewNick, const vector<CChan*>& vChans) override {
+		if (sNewNick == GetNetwork()->GetIRCSock()->GetNick()) {
 			// We are changing our own nick
 			if (Nick.NickEquals(GetNick())) {
 				// We are changing our nick away from the conf setting.
@@ -97,20 +102,20 @@ public:
 		}
 	}
 
-	void OnQuit(const CNick& Nick, const CString& sMessage, const vector<CChan*>& vChans) {
+	void OnQuit(const CNick& Nick, const CString& sMessage, const vector<CChan*>& vChans) override {
 		// If someone with the nick we want quits, be fast and get the nick
 		if (Nick.NickEquals(GetNick())) {
 			KeepNick();
 		}
 	}
 
-	void OnIRCDisconnected() {
+	void OnIRCDisconnected() override {
 		// No way we can do something if we aren't connected to IRC.
 		Disable();
 	}
 
-	void OnIRCConnected() {
-		if (!m_pNetwork->GetIRCSock()->GetNick().Equals(GetNick())) {
+	void OnIRCConnected() override {
+		if (!GetNetwork()->GetIRCSock()->GetNick().Equals(GetNick())) {
 			// We don't have the nick we want, try to get it
 			Enable();
 		}
@@ -133,9 +138,9 @@ public:
 		m_pTimer = NULL;
 	}
 
-	virtual EModRet OnUserRaw(CString& sLine) {
+	virtual EModRet OnUserRaw(CString& sLine) override {
 		// We dont care if we are not connected to IRC
-		if (!m_pNetwork->IsIRCConnected())
+		if (!GetNetwork()->IsIRCConnected())
 			return CONTINUE;
 
 		// We are trying to get the config nick and this is a /nick?
@@ -154,12 +159,12 @@ public:
 
 		// Indeed trying to change to this nick, generate a 433 for it.
 		// This way we can *always* block incoming 433s from the server.
-		PutUser(":" + m_pNetwork->GetIRCServer() + " 433 " + m_pNetwork->GetIRCNick().GetNick()
+		PutUser(":" + GetNetwork()->GetIRCServer() + " 433 " + GetNetwork()->GetIRCNick().GetNick()
 				+ " " + sNick + " :ZNC is already trying to get this nickname");
 		return CONTINUE;
 	}
 
-	virtual EModRet OnRaw(CString& sLine) {
+	virtual EModRet OnRaw(CString& sLine) override {
 		// Are we trying to get our primary nick and we caused this error?
 		// :irc.server.net 433 mynick badnick :Nickname is already in use.
 		if (m_pTimer && sLine.Token(1) == "433" && sLine.Token(3).Equals(GetNick()))
@@ -168,23 +173,21 @@ public:
 		return CONTINUE;
 	}
 
-	void OnModCommand(const CString& sCommand) {
-		CString sCmd = sCommand.AsUpper();
+	void OnEnableCommand(const CString& sCommand) {
+		Enable();
+		PutModule("Trying to get your primary nick");
+	}
 
-		if (sCmd == "ENABLE") {
-			Enable();
-			PutModule("Trying to get your primary nick");
-		} else if (sCmd == "DISABLE") {
-			Disable();
-			PutModule("No longer trying to get your primary nick");
-		} else if (sCmd == "STATE") {
-			if (m_pTimer)
-				PutModule("Currently trying to get your primary nick");
-			else
-				PutModule("Currently disabled, try 'enable'");
-		} else {
-			PutModule("Commands: Enable, Disable, State");
-		}
+	void OnDisableCommand(const CString& sCommand) {
+		Disable();
+		PutModule("No longer trying to get your primary nick");
+	}
+
+	void OnStateCommand(const CString& sCommand) {
+		if (m_pTimer)
+			PutModule("Currently trying to get your primary nick");
+		else
+			PutModule("Currently disabled, try 'enable'");
 	}
 
 private:
