@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2014 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2015 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -76,17 +76,16 @@ public:
 	MODCONSTRUCTOR(CWebAdminMod) {
 		VPair vParams;
 		vParams.push_back(make_pair("user", ""));
-		AddSubPage(new CWebSubPage("settings", "Global Settings", CWebSubPage::F_ADMIN));
-		AddSubPage(new CWebSubPage("edituser", "Your Settings", vParams));
-		AddSubPage(new CWebSubPage("traffic", "Traffic Info", CWebSubPage::F_ADMIN));
-		AddSubPage(new CWebSubPage("listusers", "List Users", CWebSubPage::F_ADMIN));
-		AddSubPage(new CWebSubPage("adduser", "Add User", CWebSubPage::F_ADMIN));
+		AddSubPage(std::make_shared<CWebSubPage>("settings", "Global Settings", CWebSubPage::F_ADMIN));
+		AddSubPage(std::make_shared<CWebSubPage>("edituser", "Your Settings", vParams));
+		AddSubPage(std::make_shared<CWebSubPage>("traffic", "Traffic Info", CWebSubPage::F_ADMIN));
+		AddSubPage(std::make_shared<CWebSubPage>("listusers", "Manage Users", CWebSubPage::F_ADMIN));
 	}
 
 	virtual ~CWebAdminMod() {
 	}
 
-	virtual bool OnLoad(const CString& sArgStr, CString& sMessage) {
+	virtual bool OnLoad(const CString& sArgStr, CString& sMessage) override {
 		if (sArgStr.empty() || CModInfo::GlobalModule != GetType())
 			return true;
 
@@ -101,6 +100,7 @@ public:
 		CString sArgs(sArgStr);
 		CString sPort;
 		CString sListenHost;
+		CString sURIPrefix;
 
 		while (sArgs.Left(1) == "-") {
 			CString sOpt = sArgs.Token(0);
@@ -150,7 +150,7 @@ public:
 		}
 
 		// Now turn that into a listener instance
-		CListener *pListener = new CListener(uPort, sListenHost, bSSL,
+		CListener *pListener = new CListener(uPort, sListenHost, sURIPrefix, bSSL,
 				(!bIPv6 ? ADDR_IPV4ONLY : ADDR_ALL), CListener::ACCEPT_HTTP);
 
 		if (!pListener->Listen()) {
@@ -164,7 +164,7 @@ public:
 	}
 
 	CUser* GetNewUser(CWebSock& WebSock, CUser* pUser) {
-		CSmartPtr<CWebSession> spSession = WebSock.GetSession();
+		std::shared_ptr<CWebSession> spSession = WebSock.GetSession();
 		CString sUsername = WebSock.GetParam("newuser");
 
 		if (sUsername.empty()) {
@@ -277,6 +277,26 @@ public:
 		pNewUser->SetTimezone(WebSock.GetParam("timezone"));
 		pNewUser->SetJoinTries(WebSock.GetParam("jointries").ToUInt());
 		pNewUser->SetMaxJoins(WebSock.GetParam("maxjoins").ToUInt());
+		pNewUser->SetAutoClearQueryBuffer(WebSock.GetParam("autoclearquerybuffer").ToBool());
+		pNewUser->SetMaxQueryBuffers(WebSock.GetParam("maxquerybuffers").ToUInt());
+
+#ifdef HAVE_ICU
+		CString sEncodingUtf = WebSock.GetParam("encoding_utf");
+		if (sEncodingUtf == "legacy") {
+			pNewUser->SetClientEncoding("");
+		}
+		CString sEncoding = WebSock.GetParam("encoding");
+		if (sEncoding.empty()) {
+			sEncoding = "UTF-8";
+		}
+		if (sEncodingUtf == "send") {
+			pNewUser->SetClientEncoding("^" + sEncoding);
+		} else if (sEncodingUtf == "receive") {
+			pNewUser->SetClientEncoding("*" + sEncoding);
+		} else if (sEncodingUtf == "simple") {
+			pNewUser->SetClientEncoding(sEncoding);
+		}
+#endif
 
 		if (spSession->IsAdmin()) {
 			pNewUser->SetDenyLoadMod(WebSock.GetParam("denyloadmod").ToBool());
@@ -398,9 +418,9 @@ public:
 		return pNetwork;
 	}
 
-	virtual CString GetWebMenuTitle() { return "webadmin"; }
-	virtual bool OnWebRequest(CWebSock& WebSock, const CString& sPageName, CTemplate& Tmpl) {
-		CSmartPtr<CWebSession> spSession = WebSock.GetSession();
+	virtual CString GetWebMenuTitle() override { return "webadmin"; }
+	virtual bool OnWebRequest(CWebSock& WebSock, const CString& sPageName, CTemplate& Tmpl) override {
+		std::shared_ptr<CWebSession> spSession = WebSock.GetSession();
 
 		if (sPageName == "settings") {
 			// Admin Check
@@ -542,7 +562,7 @@ public:
 				WebSock.PrintErrorPage("Please don't delete yourself, suicide is not the answer!");
 				return true;
 			} else if (CZNC::Get().DeleteUser(sUser)) {
-				WebSock.Redirect("listusers");
+				WebSock.Redirect(GetWebPath() + "listusers");
 				return true;
 			}
 
@@ -595,7 +615,7 @@ public:
 	}
 
 	bool ChanPage(CWebSock& WebSock, CTemplate& Tmpl, CIRCNetwork* pNetwork, CChan* pChan = NULL) {
-		CSmartPtr<CWebSession> spSession = WebSock.GetSession();
+		std::shared_ptr<CWebSession> spSession = WebSock.GetSession();
 		Tmpl.SetFile("add_edit_chan.tmpl");
 		CUser* pUser = pNetwork->GetUser();
 
@@ -682,10 +702,16 @@ public:
 			}
 		}
 
-		pChan->SetBufferCount(WebSock.GetParam("buffercount").ToUInt(), spSession->IsAdmin());
+		unsigned int uBufferCount = WebSock.GetParam("buffercount").ToUInt();
+		if (pChan->GetBufferCount() != uBufferCount) {
+			pChan->SetBufferCount(uBufferCount, spSession->IsAdmin());
+		}
 		pChan->SetDefaultModes(WebSock.GetParam("defmodes"));
 		pChan->SetInConfig(WebSock.GetParam("save").ToBool());
-		pChan->SetAutoClearChanBuffer(WebSock.GetParam("autoclearchanbuffer").ToBool());
+		bool bAutoClearChanBuffer = WebSock.GetParam("autoclearchanbuffer").ToBool();
+		if (pChan->AutoClearChanBuffer() != bAutoClearChanBuffer) {
+			pChan->SetAutoClearChanBuffer(WebSock.GetParam("autoclearchanbuffer").ToBool());
+		}
 		pChan->SetKey(WebSock.GetParam("key"));
 
 		bool bDetached = WebSock.GetParam("detached").ToBool();
@@ -716,12 +742,16 @@ public:
 			return true;
 		}
 
-		WebSock.Redirect("editnetwork?user=" + pUser->GetUserName().Escape_n(CString::EURL) + "&network=" + pNetwork->GetName().Escape_n(CString::EURL));
+		if (WebSock.HasParam("submit_return")) {
+			WebSock.Redirect(GetWebPath() + "editnetwork?user=" + pUser->GetUserName().Escape_n(CString::EURL) + "&network=" + pNetwork->GetName().Escape_n(CString::EURL));
+		} else {
+			WebSock.Redirect(GetWebPath() + "editchan?user=" + pUser->GetUserName().Escape_n(CString::EURL) + "&network=" + pNetwork->GetName().Escape_n(CString::EURL) + "&name=" + pChan->GetName().Escape_n(CString::EURL));
+		}
 		return true;
 	}
 
 	bool NetworkPage(CWebSock& WebSock, CTemplate& Tmpl, CUser* pUser, CIRCNetwork* pNetwork = NULL) {
-		CSmartPtr<CWebSession> spSession = WebSock.GetSession();
+		std::shared_ptr<CWebSession> spSession = WebSock.GetSession();
 		Tmpl.SetFile("add_edit_network.tmpl");
 
 		if (!WebSock.GetParam("submitted").ToUInt()) {
@@ -747,6 +777,14 @@ public:
 					}
 				}
 
+				// Check if module is loaded globally
+				l["CanBeLoadedGlobally"] = CString(Info.SupportsType(CModInfo::GlobalModule));
+				l["LoadedGlobally"] = CString(CZNC::Get().GetModules().FindModule(Info.GetName()) != NULL);
+
+				// Check if module is loaded by user
+				l["CanBeLoadedByUser"] = CString(Info.SupportsType(CModInfo::UserModule));
+				l["LoadedByUser"] = CString(pUser->GetModules().FindModule(Info.GetName()) != NULL);
+
 				if (!spSession->IsAdmin() && pUser->DenyLoadMod()) {
 					l["Disabled"] = "true";
 				}
@@ -795,10 +833,14 @@ public:
 				Tmpl["Ident"] = pNetwork->GetIdent();
 				Tmpl["RealName"] = pNetwork->GetRealName();
 
+				Tmpl["QuitMsg"] = pNetwork->GetQuitMsg();
+
 				Tmpl["FloodProtection"] = CString(CIRCSock::IsFloodProtected(pNetwork->GetFloodRate()));
 				Tmpl["FloodRate"] = CString(pNetwork->GetFloodRate());
 				Tmpl["FloodBurst"] = CString(pNetwork->GetFloodBurst());
 
+				Tmpl["JoinDelay"] = CString(pNetwork->GetJoinDelay());
+
 				Tmpl["IRCConnectEnabled"] = CString(pNetwork->GetIRCConnectEnabled());
 
 				const vector<CServer*>& vServers = pNetwork->GetServers();
@@ -818,13 +860,21 @@ public:
 					l["Perms"] = pChan->GetPermStr();
 					l["CurModes"] = pChan->GetModeString();
 					l["DefModes"] = pChan->GetDefaultModes();
-					l["BufferCount"] = CString(pChan->GetBufferCount());
+					if (pChan->HasBufferCountSet()) {
+						l["BufferCount"] = CString(pChan->GetBufferCount());
+					} else {
+						l["BufferCount"] = CString(pChan->GetBufferCount()) + " (default)";
+					}
 					l["Options"] = pChan->GetOptions();
 
 					if (pChan->InConfig()) {
 						l["InConfig"] = "true";
 					}
 				}
+				for (const CString& sFP : pNetwork->GetTrustedFingerprints()) {
+					CTemplate& l = Tmpl.AddRow("TrustedFingerprints");
+					l["FP"] = sFP;
+				}
 			} else {
 				if (!spSession->IsAdmin() && !pUser->HasSpaceForNewNetwork()) {
 					WebSock.PrintErrorPage("Network number limit reached. Ask an admin to increase the limit for you, or delete unneeded networks from Your Settings.");
@@ -837,6 +887,7 @@ public:
 				Tmpl["FloodProtection"] = "true";
 				Tmpl["FloodRate"] = "1.0";
 				Tmpl["FloodBurst"] = "4";
+				Tmpl["JoinDelay"] = "0";
 			}
 
 			FOR_EACH_MODULE(i, make_pair(pUser, pNetwork)) {
@@ -849,30 +900,57 @@ public:
 				}
 			}
 
+#ifdef HAVE_ICU
+			for (const CString& sEncoding : CUtils::GetEncodings()) {
+				CTemplate& l = Tmpl.AddRow("EncodingLoop");
+				l["Encoding"] = sEncoding;
+			}
+			const CString sEncoding = pNetwork ? pNetwork->GetEncoding() : "^UTF-8";
+			if (sEncoding.empty()) {
+				Tmpl["EncodingUtf"] = "legacy";
+			} else if (sEncoding[0] == '*') {
+				Tmpl["EncodingUtf"] = "receive";
+				Tmpl["Encoding"] = sEncoding.substr(1);
+			} else if (sEncoding[0] == '^') {
+				Tmpl["EncodingUtf"] = "send";
+				Tmpl["Encoding"] = sEncoding.substr(1);
+			} else {
+				Tmpl["EncodingUtf"] = "simple";
+				Tmpl["Encoding"] = sEncoding;
+			}
+#else
+			Tmpl["EncodingDisabled"] = "true";
+			Tmpl["EncodingUtf"] = "legacy";
+#endif
+
 			return true;
 		}
 
-		CString sName = WebSock.GetParam("network").Trim_n();
+		CString sName = WebSock.GetParam("name").Trim_n();
 		if (sName.empty()) {
 			WebSock.PrintErrorPage("Network name is a required argument");
 			return true;
 		}
-
-		if (!pNetwork) {
-			if (!spSession->IsAdmin() && !pUser->HasSpaceForNewNetwork()) {
-				WebSock.PrintErrorPage("Network number limit reached. Ask an admin to increase the limit for you, or delete few old ones from Your Settings");
-				return true;
-			}
-			if (!CIRCNetwork::IsValidNetwork(sName)) {
-				WebSock.PrintErrorPage("Network name should be alphanumeric");
-				return true;
-			}
+		if (!pNetwork && !spSession->IsAdmin() && !pUser->HasSpaceForNewNetwork()) {
+			WebSock.PrintErrorPage("Network number limit reached. Ask an admin to increase the limit for you, or delete few old ones from Your Settings");
+			return true;
+		}
+		if (!pNetwork || pNetwork->GetName() != sName) {
 			CString sNetworkAddError;
+			CIRCNetwork* pOldNetwork = pNetwork;
 			pNetwork = pUser->AddNetwork(sName, sNetworkAddError);
 			if (!pNetwork) {
 				WebSock.PrintErrorPage(sNetworkAddError);
 				return true;
 			}
+			if (pOldNetwork) {
+				for (CModule* pModule : pOldNetwork->GetModules()) {
+					CString sPath = pUser->GetUserPath() + "/networks/" + sName + "/moddata/" + pModule->GetModName();
+					pModule->MoveRegistry(sPath);
+				}
+				pNetwork->Clone(*pOldNetwork, false);
+				pUser->DeleteNetwork(pOldNetwork->GetName());
+			}
 		}
 
 		CString sArg;
@@ -882,6 +960,8 @@ public:
 		pNetwork->SetIdent(WebSock.GetParam("ident"));
 		pNetwork->SetRealName(WebSock.GetParam("realname"));
 
+		pNetwork->SetQuitMsg(WebSock.GetParam("quitmsg"));
+
 		pNetwork->SetIRCConnectEnabled(WebSock.GetParam("doconnect").ToBool());
 
 		sArg = WebSock.GetParam("bindhost");
@@ -914,6 +994,26 @@ public:
 			pNetwork->SetFloodRate(-1);
 		}
 
+		pNetwork->SetJoinDelay(WebSock.GetParam("joindelay").ToUShort());
+
+#ifdef HAVE_ICU
+		CString sEncodingUtf = WebSock.GetParam("encoding_utf");
+		if (sEncodingUtf == "legacy") {
+			pNetwork->SetEncoding("");
+		}
+		CString sEncoding = WebSock.GetParam("encoding");
+		if (sEncoding.empty()) {
+			sEncoding = "UTF-8";
+		}
+		if (sEncodingUtf == "send") {
+			pNetwork->SetEncoding("^" + sEncoding);
+		} else if (sEncodingUtf == "receive") {
+			pNetwork->SetEncoding("*" + sEncoding);
+		} else if (sEncodingUtf == "simple") {
+			pNetwork->SetEncoding(sEncoding);
+		}
+#endif
+
 		VCString vsArgs;
 
 		pNetwork->DelServers();
@@ -922,6 +1022,14 @@ public:
 			pNetwork->AddServer(vsArgs[a].Trim_n());
 		}
 
+		WebSock.GetRawParam("fingerprints").Split("\n", vsArgs);
+		while (!pNetwork->GetTrustedFingerprints().empty()) {
+			pNetwork->DelTrustedFingerprint(*pNetwork->GetTrustedFingerprints().begin());
+		}
+		for (const CString& sFP : vsArgs) {
+			pNetwork->AddTrustedFingerprint(sFP);
+		}
+
 		WebSock.GetParamValues("channel", vsArgs);
 		for (unsigned int a = 0; a < vsArgs.size(); a++) {
 			const CString& sChan = vsArgs[a];
@@ -990,7 +1098,11 @@ public:
 			return true;
 		}
 
-		WebSock.Redirect("edituser?user=" + pUser->GetUserName().Escape_n(CString::EURL));
+		if (WebSock.HasParam("submit_return")) {
+			WebSock.Redirect(GetWebPath() + "edituser?user=" + pUser->GetUserName().Escape_n(CString::EURL));
+		} else {
+			WebSock.Redirect(GetWebPath() + "editnetwork?user=" + pUser->GetUserName().Escape_n(CString::EURL) + "&network=" + pNetwork->GetName().Escape_n(CString::EURL));
+		}
 		return true;
 	}
 
@@ -1026,7 +1138,7 @@ public:
 			return true;
 		}
 
-		WebSock.Redirect("edituser?user=" + pUser->GetUserName().Escape_n(CString::EURL));
+		WebSock.Redirect(GetWebPath() + "edituser?user=" + pUser->GetUserName().Escape_n(CString::EURL));
 		return false;
 	}
 
@@ -1046,17 +1158,15 @@ public:
 			return true;
 		}
 
-		WebSock.Redirect("editnetwork?user=" + pNetwork->GetUser()->GetUserName().Escape_n(CString::EURL) + "&network=" + pNetwork->GetName().Escape_n(CString::EURL));
+		WebSock.Redirect(GetWebPath() + "editnetwork?user=" + pNetwork->GetUser()->GetUserName().Escape_n(CString::EURL) + "&network=" + pNetwork->GetName().Escape_n(CString::EURL));
 		return false;
 	}
 
 	bool UserPage(CWebSock& WebSock, CTemplate& Tmpl, CUser* pUser = NULL) {
-		CSmartPtr<CWebSession> spSession = WebSock.GetSession();
+		std::shared_ptr<CWebSession> spSession = WebSock.GetSession();
 		Tmpl.SetFile("add_edit_user.tmpl");
 
 		if (!WebSock.GetParam("submitted").ToUInt()) {
-			CString sAllowedHosts, sServers, sChans, sCTCPReplies;
-
 			if (pUser) {
 				Tmpl["Action"] = "edituser";
 				Tmpl["Title"] = "Edit User [" + pUser->GetUserName() + "]";
@@ -1089,6 +1199,7 @@ public:
 				Tmpl["JoinTries"] = CString(pUser->JoinTries());
 				Tmpl["MaxNetworks"] = CString(pUser->MaxNetworks());
 				Tmpl["MaxJoins"] = CString(pUser->MaxJoins());
+				Tmpl["MaxQueryBuffers"] = CString(pUser->MaxQueryBuffers());
 
 				const set<CString>& ssAllowedHosts = pUser->GetAllowedHosts();
 				for (set<CString>::const_iterator it = ssAllowedHosts.begin(); it != ssAllowedHosts.end(); ++it) {
@@ -1126,6 +1237,29 @@ public:
 				l["TZ"] = *i;
 			}
 
+#ifdef HAVE_ICU
+			for (const CString& sEncoding : CUtils::GetEncodings()) {
+				CTemplate& l = Tmpl.AddRow("EncodingLoop");
+				l["Encoding"] = sEncoding;
+			}
+			const CString sEncoding = pUser ? pUser->GetClientEncoding() : "^UTF-8";
+			if (sEncoding.empty()) {
+				Tmpl["EncodingUtf"] = "legacy";
+			} else if (sEncoding[0] == '*') {
+				Tmpl["EncodingUtf"] = "receive";
+				Tmpl["Encoding"] = sEncoding.substr(1);
+			} else if (sEncoding[0] == '^') {
+				Tmpl["EncodingUtf"] = "send";
+				Tmpl["Encoding"] = sEncoding.substr(1);
+			} else {
+				Tmpl["EncodingUtf"] = "simple";
+				Tmpl["Encoding"] = sEncoding;
+			}
+#else
+			Tmpl["EncodingDisabled"] = "true";
+			Tmpl["EncodingUtf"] = "legacy";
+#endif
+
 			// To change BindHosts be admin or don't have DenySetBindHost
 			if (spSession->IsAdmin() || !spSession->GetUser()->DenySetBindHost()) {
 				Tmpl["BindHostEdit"] = "true";
@@ -1200,8 +1334,22 @@ public:
 				l["ArgsHelpText"] = Info.GetArgsHelpText();
 
 				CModule *pModule = NULL;
-				if (pUser)
+				if (pUser) {
 					pModule = pUser->GetModules().FindModule(Info.GetName());
+					// Check if module is loaded by all or some networks
+					const vector<CIRCNetwork*>& userNetworks = pUser->GetNetworks();
+					unsigned int networksWithRenderedModuleCount = 0;
+					for (unsigned int networkIndex = 0; networkIndex < userNetworks.size(); ++networkIndex) {
+						const CIRCNetwork* pCurrentNetwork = userNetworks[networkIndex];
+						const CModules& networkModules = pCurrentNetwork->GetModules();
+						if (networkModules.FindModule(Info.GetName())) {
+							networksWithRenderedModuleCount++;
+						}
+					}
+					l["CanBeLoadedByNetwork"] = CString(Info.SupportsType(CModInfo::NetworkModule));
+					l["LoadedByAllNetworks"] = CString(networksWithRenderedModuleCount == userNetworks.size());
+					l["LoadedBySomeNetworks"] = CString(networksWithRenderedModuleCount != 0);
+				}
 				if (pModule) {
 					l["Checked"] = "true";
 					l["Args"] = pModule->GetArgs();
@@ -1209,6 +1357,9 @@ public:
 						l["Disabled"] = "true";
 					}
 				}
+				l["CanBeLoadedGlobally"] = CString(Info.SupportsType(CModInfo::GlobalModule));
+				// Check if module is loaded globally
+				l["LoadedGlobally"] = CString(CZNC::Get().GetModules().FindModule(Info.GetName()) != NULL);
 
 				if (!spSession->IsAdmin() && pUser && pUser->DenyLoadMod()) {
 					l["Disabled"] = "true";
@@ -1256,6 +1407,12 @@ public:
 				if (pUser && pUser->DenySetBindHost()) { o11["Checked"] = "true"; }
 			}
 
+			CTemplate& o12 = Tmpl.AddRow("OptionLoop");
+			o12["Name"] = "autoclearquerybuffer";
+			o12["DisplayName"] = "Auto Clear Query Buffer";
+			o12["Tooltip"] = "Automatically Clear Query Buffer After Playback";
+			if (!pUser || pUser->AutoClearQueryBuffer()) { o12["Checked"] = "true"; }
+
 			FOR_EACH_MODULE(i, pUser) {
 				CTemplate& mod = Tmpl.AddRow("EmbeddedModuleLoop");
 				mod.insert(Tmpl.begin(), Tmpl.end());
@@ -1325,10 +1482,10 @@ public:
 			return true;
 		}
 
-		if (!spSession->IsAdmin()) {
-			WebSock.Redirect("edituser");
+		if (spSession->IsAdmin() && WebSock.HasParam("submit_return")) {
+			WebSock.Redirect(GetWebPath() + "listusers");
 		} else {
-			WebSock.Redirect("listusers");
+			WebSock.Redirect(GetWebPath() + "edituser?user=" + pUser->GetUserName());
 		}
 
 		/* we don't want the template to be printed while we redirect */
@@ -1336,9 +1493,9 @@ public:
 	}
 
 	bool ListUsersPage(CWebSock& WebSock, CTemplate& Tmpl) {
-		CSmartPtr<CWebSession> spSession = WebSock.GetSession();
+		std::shared_ptr<CWebSession> spSession = WebSock.GetSession();
 		const map<CString,CUser*>& msUsers = CZNC::Get().GetUserMap();
-		Tmpl["Title"] = "List Users";
+		Tmpl["Title"] = "Manage Users";
 		Tmpl["Action"] = "listusers";
 
 		unsigned int a = 0;
@@ -1360,7 +1517,7 @@ public:
 	}
 
 	bool TrafficPage(CWebSock& WebSock, CTemplate& Tmpl) {
-		CSmartPtr<CWebSession> spSession = WebSock.GetSession();
+		Tmpl["Title"] = "Traffic Info";
 		Tmpl["Uptime"] = CZNC::Get().GetUptime();
 
 		const map<CString,CUser*>& msUsers = CZNC::Get().GetUserMap();
@@ -1426,6 +1583,7 @@ public:
 	bool AddListener(CWebSock& WebSock, CTemplate& Tmpl) {
 		unsigned short uPort = WebSock.GetParam("port").ToUShort();
 		CString sHost = WebSock.GetParam("host");
+		CString sURIPrefix = WebSock.GetParam("uriprefix");
 		if (sHost == "*") sHost = "";
 		bool bSSL = WebSock.GetParam("ssl").ToBool();
 		bool bIPv4 = WebSock.GetParam("ipv4").ToBool();
@@ -1466,7 +1624,7 @@ public:
 		}
 
 		CString sMessage;
-		if (CZNC::Get().AddListener(uPort, sHost, bSSL, eAddr, eAccept, sMessage)) {
+		if (CZNC::Get().AddListener(uPort, sHost, sURIPrefix, bSSL, eAddr, eAccept, sMessage)) {
 			if (!sMessage.empty()) {
 				WebSock.GetSession()->AddSuccess(sMessage);
 			}
@@ -1518,7 +1676,6 @@ public:
 	bool SettingsPage(CWebSock& WebSock, CTemplate& Tmpl) {
 		Tmpl.SetFile("settings.tmpl");
 		if (!WebSock.GetParam("submitted").ToUInt()) {
-			CString sBindHosts, sMotd;
 			Tmpl["Action"] = "settings";
 			Tmpl["Title"] = "Settings";
 			Tmpl["StatusPrefix"] = CZNC::Get().GetStatusPrefix();
@@ -1527,6 +1684,7 @@ public:
 			Tmpl["ServerThrottle"] = CString(CZNC::Get().GetServerThrottle());
 			Tmpl["AnonIPLimit"] = CString(CZNC::Get().GetAnonIPLimit());
 			Tmpl["ProtectWebSessions"] = CString(CZNC::Get().GetProtectWebSessions());
+			Tmpl["HideVersion"] = CString(CZNC::Get().GetHideVersion());
 
 			const VCString& vsBindHosts = CZNC::Get().GetBindHosts();
 			for (unsigned int a = 0; a < vsBindHosts.size(); a++) {
@@ -1551,6 +1709,8 @@ public:
 				l["IsWeb"] = CString(pListener->GetAcceptType() != CListener::ACCEPT_IRC);
 				l["IsIRC"] = CString(pListener->GetAcceptType() != CListener::ACCEPT_HTTP);
 
+				l["URIPrefix"] = pListener->GetURIPrefix() + "/";
+
 				// simple protection for user from shooting his own foot
 				// TODO check also for hosts/families
 				// such check is only here, user still can forge HTTP request to delete web port
@@ -1614,6 +1774,38 @@ public:
 				l["Wiki"] = Info.GetWikiPage();
 				l["HasArgs"] = CString(Info.GetHasArgs());
 				l["ArgsHelpText"] = Info.GetArgsHelpText();
+
+				// Check if the module is loaded by all or some users, and/or by all or some networks
+				unsigned int usersWithRenderedModuleCount = 0;
+				unsigned int networksWithRenderedModuleCount = 0;
+				unsigned int networksCount = 0;
+				const map<CString,CUser*>& allUsers = CZNC::Get().GetUserMap();
+				for (map<CString,CUser*>::const_iterator usersIt = allUsers.begin(); usersIt != allUsers.end(); ++usersIt) {
+					const CUser& User = *usersIt->second;
+					
+					// Count users which has loaded a render module
+					const CModules& userModules = User.GetModules();
+					if (userModules.FindModule(Info.GetName())) {
+						usersWithRenderedModuleCount++;
+					}
+					// Count networks which has loaded a render module
+					const vector<CIRCNetwork*>& userNetworks = User.GetNetworks();
+					networksCount += userNetworks.size();
+					for (unsigned int networkIndex = 0; networkIndex < userNetworks.size(); ++networkIndex)
+					{
+						const CIRCNetwork *pCurrentNetwork = userNetworks[networkIndex];
+						if (pCurrentNetwork->GetModules().FindModule(Info.GetName())) {
+							networksWithRenderedModuleCount++;
+						}
+					}
+				}
+				l["CanBeLoadedByNetwork"] = CString(Info.SupportsType(CModInfo::NetworkModule));
+				l["LoadedByAllNetworks"] = CString(networksWithRenderedModuleCount == networksCount);
+				l["LoadedBySomeNetworks"] = CString(networksWithRenderedModuleCount != 0);
+
+				l["CanBeLoadedByUser"] = CString(Info.SupportsType(CModInfo::UserModule));
+				l["LoadedByAllUsers"] = CString(usersWithRenderedModuleCount == allUsers.size());
+				l["LoadedBySomeUsers"] = CString(usersWithRenderedModuleCount != 0);
 			}
 
 			return true;
@@ -1626,6 +1818,7 @@ public:
 		sArg = WebSock.GetParam("serverthrottle"); CZNC::Get().SetServerThrottle(sArg.ToUInt());
 		sArg = WebSock.GetParam("anoniplimit"); CZNC::Get().SetAnonIPLimit(sArg.ToUInt());
 		sArg = WebSock.GetParam("protectwebsessions"); CZNC::Get().SetProtectWebSessions(sArg.ToBool());
+		sArg = WebSock.GetParam("hideversion"); CZNC::Get().SetHideVersion(sArg.ToBool());
 
 		VCString vsArgs;
 		WebSock.GetRawParam("motd").Split("\n", vsArgs);
@@ -1694,7 +1887,7 @@ public:
 			WebSock.GetSession()->AddError("Settings changed, but config was not written");
 		}
 
-		WebSock.Redirect("settings");
+		WebSock.Redirect(GetWebPath() + "settings");
 		/* we don't want the template to be printed while we redirect */
 		return false;
 	}
@@ -1705,4 +1898,4 @@ template<> void TModInfo<CWebAdminMod>(CModInfo& Info) {
 	Info.SetWikiPage("webadmin");
 }
 
-GLOBALMODULEDEFS(CWebAdminMod, "Web based administration module")
+GLOBALMODULEDEFS(CWebAdminMod, "Web based administration module.")
