@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2014 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2015 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -28,11 +28,11 @@ using std::set;
 
 #define MAX_POST_SIZE	1024 * 1024
 
-CHTTPSock::CHTTPSock(CModule *pMod) : CSocket(pMod) {
+CHTTPSock::CHTTPSock(CModule *pMod, const CString& sURIPrefix) : CSocket(pMod), m_sURIPrefix(sURIPrefix) {
 	Init();
 }
 
-CHTTPSock::CHTTPSock(CModule *pMod, const CString& sHostname, unsigned short uPort, int iTimeout) : CSocket(pMod, sHostname, uPort, iTimeout) {
+CHTTPSock::CHTTPSock(CModule *pMod, const CString& sURIPrefix, const CString& sHostname, unsigned short uPort, int iTimeout) : CSocket(pMod, sHostname, uPort, iTimeout), m_sURIPrefix(sURIPrefix) {
 	Init();
 }
 
@@ -122,11 +122,43 @@ void CHTTPSock::ReadLine(const CString& sData) {
 		sLine.Token(2).Base64Decode(sUnhashed);
 		m_sUser = sUnhashed.Token(0, false, ":");
 		m_sPass = sUnhashed.Token(1, true, ":");
-		m_bLoggedIn = OnLogin(m_sUser, m_sPass);
+		m_bLoggedIn = OnLogin(m_sUser, m_sPass, true);
 	} else if (sName.Equals("Content-Length:")) {
 		m_uPostLen = sLine.Token(1).ToULong();
 		if (m_uPostLen > MAX_POST_SIZE)
 			PrintErrorPage(413, "Request Entity Too Large", "The request you sent was too large.");
+	} else if (sName.Equals("X-Forwarded-For:")) {
+		// X-Forwarded-For: client, proxy1, proxy2
+		if (m_sForwardedIP.empty()) {
+			const VCString& vsTrustedProxies = CZNC::Get().GetTrustedProxies();
+			CString sIP = GetRemoteIP();
+
+			VCString vsIPs;
+			sLine.Token(1, true).Split(",", vsIPs, false, "", "", false, true);
+
+			while (!vsIPs.empty()) {
+				// sIP told us that it got connection from vsIPs.back()
+				// check if sIP is trusted proxy
+				bool bTrusted = false;
+				for (VCString::const_iterator it = vsTrustedProxies.begin(); it != vsTrustedProxies.end(); ++it) {
+					if (sIP.WildCmp(*it)) {
+						bTrusted = true;
+						break;
+					}
+				}
+				if (bTrusted) {
+					// sIP is trusted proxy, so use vsIPs.back() as new sIP
+					sIP = vsIPs.back();
+					vsIPs.pop_back();
+				} else {
+					break;
+				}
+			}
+
+			// either sIP is not trusted proxy, or it's in the beginning of the X-Forwarded-For list
+			// in both cases use it as the endpoind
+			m_sForwardedIP = sIP;
+		}
 	} else if (sName.Equals("If-None-Match:")) {
 		// this is for proper client cache support (HTTP 304) on static files:
 		m_sIfNoneMatch = sLine.Token(1, true);
@@ -149,6 +181,14 @@ void CHTTPSock::ReadLine(const CString& sData) {
 	}
 }
 
+CString CHTTPSock::GetRemoteIP() const {
+	if (!m_sForwardedIP.empty()) {
+		return m_sForwardedIP;
+	}
+
+	return CSocket::GetRemoteIP();
+}
+
 CString CHTTPSock::GetDate(time_t stamp) {
 	struct tm tm;
 	std::stringstream stream;
@@ -174,7 +214,16 @@ CString CHTTPSock::GetDate(time_t stamp) {
 void CHTTPSock::GetPage() {
 	DEBUG("Page Request [" << m_sURI << "] ");
 
-	OnPageRequest(m_sURI);
+	// Check that the requested path starts with the prefix. Strip it if so.
+	if (!m_sURI.TrimPrefix(m_sURIPrefix)) {
+		DEBUG("INVALID path => Does not start with prefix [" + m_sURIPrefix + "]");
+		DEBUG("Expected prefix:   " << m_sURIPrefix);
+		DEBUG("Requested path:    " << m_sURI);
+		Redirect("/");
+	} else {
+		OnPageRequest(m_sURI);
+	}
+
 }
 
 #ifdef HAVE_ZLIB
@@ -221,6 +270,7 @@ void CHTTPSock::PrintPage(const CString& sPage) {
 			} while(zStatus == Z_OK);
 
 			Close(Csock::CLT_AFTERWRITE);
+			deflateEnd(&zStrm);
 			return;
 		}
 
@@ -455,6 +505,10 @@ const CString& CHTTPSock::GetParamString() const {
 	return m_sPostData;
 }
 
+const CString& CHTTPSock::GetURIPrefix() const {
+	return m_sURIPrefix;
+}
+
 bool CHTTPSock::HasParam(const CString& sName, bool bPost) const {
 	if (bPost)
 		return (m_msvsPOSTParams.find(sName) != m_msvsPOSTParams.end());
@@ -610,7 +664,7 @@ bool CHTTPSock::ForceLogin() {
 	return false;
 }
 
-bool CHTTPSock::OnLogin(const CString& sUser, const CString& sPass) {
+bool CHTTPSock::OnLogin(const CString& sUser, const CString& sPass, bool bBasic) {
 	return false;
 }
 
@@ -672,13 +726,19 @@ bool CHTTPSock::Redirect(const CString& sURL) {
 	if (SentHeader()) {
 		DEBUG("Redirect() - Header was already sent");
 		return false;
+	} else if(!sURL.StartsWith("/")) {
+		// HTTP/1.1 only admits absolute URIs for the Location header.
+		DEBUG("Redirect to relative URI [" + sURL + "] is not allowed.");
+		return false;
+	} else {
+		CString location = m_sURIPrefix + sURL;
+
+		DEBUG("- Redirect to [" << location << "] with prefix [" + m_sURIPrefix + "]");
+		AddHeader("Location", location);
+		PrintErrorPage(302, "Found", "The document has moved <a href=\"" + location.Escape_n(CString::EHTML) + "\">here</a>.");
+
+		return true;
 	}
-
-	DEBUG("- Redirect to [" << sURL << "]");
-	AddHeader("Location", sURL);
-	PrintErrorPage(302, "Found", "The document has moved <a href=\"" + sURL.Escape_n(CString::EHTML) + "\">here</a>.");
-
-	return true;
 }
 
 void CHTTPSock::Connected() {
