@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2014 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2015 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,10 +21,74 @@
 #include <znc/IRCSock.h>
 #include <znc/Server.h>
 #include <znc/Chan.h>
+#include <znc/Query.h>
+#include <algorithm>
 
 using std::vector;
 using std::set;
 
+class CIRCNetworkPingTimer : public CCron {
+public:
+	CIRCNetworkPingTimer(CIRCNetwork *pNetwork) : CCron() {
+		m_pNetwork = pNetwork;
+		SetName("CIRCNetworkPingTimer::" + m_pNetwork->GetUser()->GetUserName() + "::" + m_pNetwork->GetName());
+		Start(CIRCNetwork::PING_SLACK);
+	}
+
+	virtual ~CIRCNetworkPingTimer() {}
+
+protected:
+	virtual void RunJob() {
+		CIRCSock* pIRCSock = m_pNetwork->GetIRCSock();
+
+		if (pIRCSock && pIRCSock->GetTimeSinceLastDataTransaction() >= CIRCNetwork::PING_FREQUENCY) {
+			pIRCSock->PutIRC("PING :ZNC");
+		}
+
+		const vector<CClient*>& vClients = m_pNetwork->GetClients();
+		for (size_t b = 0; b < vClients.size(); b++) {
+			CClient* pClient = vClients[b];
+
+			if (pClient->GetTimeSinceLastDataTransaction() >= CIRCNetwork::PING_FREQUENCY) {
+				pClient->PutClient("PING :ZNC");
+			}
+		}
+	}
+
+private:
+	CIRCNetwork* m_pNetwork;
+};
+
+class CIRCNetworkJoinTimer : public CCron {
+public:
+	CIRCNetworkJoinTimer(CIRCNetwork *pNetwork) : CCron(), m_bDelayed(false), m_pNetwork(pNetwork) {
+		SetName("CIRCNetworkJoinTimer::" + m_pNetwork->GetUser()->GetUserName() + "::" + m_pNetwork->GetName());
+		Start(CIRCNetwork::JOIN_FREQUENCY);
+	}
+
+	virtual ~CIRCNetworkJoinTimer() {}
+
+	void Delay(unsigned short int uDelay) {
+		m_bDelayed = true;
+		Start(uDelay);
+	}
+
+protected:
+	virtual void RunJob() {
+		if (m_bDelayed) {
+			m_bDelayed = false;
+			Start(CIRCNetwork::JOIN_FREQUENCY);
+		}
+		if (m_pNetwork->IsIRCConnected()) {
+			m_pNetwork->JoinChans();
+		}
+	}
+
+private:
+	bool         m_bDelayed;
+	CIRCNetwork* m_pNetwork;
+};
+
 bool CIRCNetwork::IsValidNetwork(const CString& sNetwork) {
 	// ^[-\w]+$
 
@@ -56,13 +120,22 @@ CIRCNetwork::CIRCNetwork(CUser *pUser, const CString& sName) {
 
 	m_sChanPrefixes = "";
 	m_bIRCAway = false;
+	m_sEncoding = "";
 
 	m_fFloodRate = 1;
 	m_uFloodBurst = 4;
 
+	m_uJoinDelay = 0;
+
 	m_RawBuffer.SetLineCount(100, true);   // This should be more than enough raws, especially since we are buffering the MOTD separately
 	m_MotdBuffer.SetLineCount(200, true);  // This should be more than enough motd lines
-	m_QueryBuffer.SetLineCount(250, true);
+	m_NoticeBuffer.SetLineCount(250, true);
+
+	m_pPingTimer = new CIRCNetworkPingTimer(this);
+	CZNC::Get().GetManager().AddCron(m_pPingTimer);
+
+	m_pJoinTimer = new CIRCNetworkJoinTimer(this);
+	CZNC::Get().GetManager().AddCron(m_pJoinTimer);
 
 	SetIRCConnectEnabled(true);
 }
@@ -78,10 +151,11 @@ CIRCNetwork::CIRCNetwork(CUser *pUser, const CIRCNetwork &Network) {
 
 	m_sChanPrefixes = "";
 	m_bIRCAway = false;
+	m_sEncoding = "";
 
 	m_RawBuffer.SetLineCount(100, true);   // This should be more than enough raws, especially since we are buffering the MOTD separately
 	m_MotdBuffer.SetLineCount(200, true);  // This should be more than enough motd lines
-	m_QueryBuffer.SetLineCount(250, true);
+	m_NoticeBuffer.SetLineCount(250, true);
 
 	Clone(Network);
 }
@@ -93,12 +167,16 @@ void CIRCNetwork::Clone(const CIRCNetwork& Network, bool bCloneName) {
 
 	m_fFloodRate = Network.GetFloodRate();
 	m_uFloodBurst = Network.GetFloodBurst();
+	m_uJoinDelay = Network.GetJoinDelay();
 
 	SetNick(Network.GetNick());
 	SetAltNick(Network.GetAltNick());
 	SetIdent(Network.GetIdent());
 	SetRealName(Network.GetRealName());
 	SetBindHost(Network.GetBindHost());
+	SetEncoding(Network.GetEncoding());
+	SetQuitMsg(Network.GetQuitMsg());
+	m_ssTrustedFingerprints = Network.m_ssTrustedFingerprints;
 
 	// Servers
 	const vector<CServer*>& vServers = Network.GetServers();
@@ -219,10 +297,19 @@ CIRCNetwork::~CIRCNetwork() {
 	}
 	m_vChans.clear();
 
+	// Delete Queries
+	for (vector<CQuery*>::const_iterator it = m_vQueries.begin(); it != m_vQueries.end(); ++it) {
+		delete *it;
+	}
+	m_vQueries.clear();
+
 	SetUser(NULL);
 
 	// Make sure we are not in the connection queue
 	CZNC::Get().GetConnectionQueue().remove(this);
+
+	CZNC::Get().GetManager().DelCronByAddr(m_pPingTimer);
+	CZNC::Get().GetManager().DelCronByAddr(m_pJoinTimer);
 }
 
 void CIRCNetwork::DelServers() {
@@ -232,7 +319,7 @@ void CIRCNetwork::DelServers() {
 	m_vServers.clear();
 }
 
-CString CIRCNetwork::GetNetworkPath() {
+CString CIRCNetwork::GetNetworkPath() const {
 	CString sNetworkPath = m_pUser->GetUserPath() + "/networks/" + m_sName;
 
 	if (!CFile::Exists(sNetworkPath)) {
@@ -259,6 +346,8 @@ bool CIRCNetwork::ParseConfig(CConfig *pConfig, CString& sError, bool bUpgrade)
 			{ "ident", &CIRCNetwork::SetIdent },
 			{ "realname", &CIRCNetwork::SetRealName },
 			{ "bindhost", &CIRCNetwork::SetBindHost },
+			{ "encoding", &CIRCNetwork::SetEncoding },
+			{ "quitmsg", &CIRCNetwork::SetQuitMsg },
 		};
 		size_t numStringOptions = sizeof(StringOptions) / sizeof(StringOptions[0]);
 		TOption<bool> BoolOptions[] = {
@@ -271,6 +360,7 @@ bool CIRCNetwork::ParseConfig(CConfig *pConfig, CString& sError, bool bUpgrade)
 		size_t numDoubleOptions = sizeof(DoubleOptions) / sizeof(DoubleOptions[0]);
 		TOption<short unsigned int> SUIntOptions[] = {
 			{ "floodburst", &CIRCNetwork::SetFloodBurst },
+			{ "joindelay", &CIRCNetwork::SetJoinDelay },
 		};
 		size_t numSUIntOptions = sizeof(SUIntOptions) / sizeof(SUIntOptions[0]);
 
@@ -302,38 +392,46 @@ bool CIRCNetwork::ParseConfig(CConfig *pConfig, CString& sError, bool bUpgrade)
 		for (vit = vsList.begin(); vit != vsList.end(); ++vit) {
 			CString sValue = *vit;
 			CString sModName = sValue.Token(0);
+			CString sNotice = "Loading network module [" + sModName + "]";
 
 			// XXX Legacy crap, added in ZNC 0.203, modified in 0.207
 			// Note that 0.203 == 0.207
 			if (sModName == "away") {
-				CUtils::PrintMessage("NOTICE: [away] was renamed, "
-						"loading [awaystore] instead");
+				sNotice = "NOTICE: [away] was renamed, loading [awaystore] instead";
 				sModName = "awaystore";
 			}
 
 			// XXX Legacy crap, added in ZNC 0.207
 			if (sModName == "autoaway") {
-				CUtils::PrintMessage("NOTICE: [autoaway] was renamed, "
-						"loading [awaystore] instead");
+				sNotice = "NOTICE: [autoaway] was renamed, loading [awaystore] instead";
 				sModName = "awaystore";
 			}
 		
 			// XXX Legacy crap, added in 1.1; fakeonline module was dropped in 1.0 and returned in 1.1
 			if (sModName == "fakeonline") {
-				CUtils::PrintMessage("NOTICE: [fakeonline] was renamed, loading [modules_online] instead");
+				sNotice = "NOTICE: [fakeonline] was renamed, loading [modules_online] instead";
 				sModName = "modules_online";
 			}
 
-			CUtils::PrintAction("Loading network module [" + sModName + "]");
 			CString sModRet;
 			CString sArgs = sValue.Token(1, true);
 
-			bool bModRet = GetModules().LoadModule(sModName, sArgs, CModInfo::NetworkModule, GetUser(), this, sModRet);
+			bool bModRet = LoadModule(sModName, sArgs, sNotice, sModRet);
 
-			CUtils::PrintStatus(bModRet, sModRet);
 			if (!bModRet) {
-				sError = sModRet;
-				return false;
+				// XXX The awaynick module was retired in 1.6 (still available as external module)
+				if (sModName == "awaynick") {
+					// load simple_away instead, unless it's already on the list
+					if (std::find(vsList.begin(), vsList.end(), "simple_away") == vsList.end()) {
+						sNotice = "Loading network module [simple_away] instead";
+						sModName = "simple_away";
+						// not a fatal error if simple_away is not available
+						LoadModule(sModName, sArgs, sNotice, sModRet);
+					}
+				} else {
+					sError = sModRet;
+					return false;
+				}
 			}
 		}
 	}
@@ -344,6 +442,11 @@ bool CIRCNetwork::ParseConfig(CConfig *pConfig, CString& sError, bool bUpgrade)
 		CUtils::PrintStatus(AddServer(*vit));
 	}
 
+	pConfig->FindStringVector("trustedserverfingerprint", vsList);
+	for (const CString& sFP : vsList) {
+		AddTrustedFingerprint(sFP);
+	}
+
 	pConfig->FindStringVector("chan", vsList);
 	for (vit = vsList.begin(); vit != vsList.end(); ++vit) {
 		AddChan(*vit, true);
@@ -380,7 +483,7 @@ bool CIRCNetwork::ParseConfig(CConfig *pConfig, CString& sError, bool bUpgrade)
 	return true;
 }
 
-CConfig CIRCNetwork::ToConfig() {
+CConfig CIRCNetwork::ToConfig() const {
 	CConfig config;
 
 	if (!m_sNick.empty()) {
@@ -405,9 +508,15 @@ CConfig CIRCNetwork::ToConfig() {
 	config.AddKeyValuePair("IRCConnectEnabled", CString(GetIRCConnectEnabled()));
 	config.AddKeyValuePair("FloodRate", CString(GetFloodRate()));
 	config.AddKeyValuePair("FloodBurst", CString(GetFloodBurst()));
+	config.AddKeyValuePair("JoinDelay", CString(GetJoinDelay()));
+	config.AddKeyValuePair("Encoding", m_sEncoding);
+
+	if (!m_sQuitMsg.empty()) {
+		config.AddKeyValuePair("QuitMsg", m_sQuitMsg);
+	}
 
 	// Modules
-	CModules& Mods = GetModules();
+	const CModules& Mods = GetModules();
 
 	if (!Mods.empty()) {
 		for (unsigned int a = 0; a < Mods.size(); a++) {
@@ -426,6 +535,10 @@ CConfig CIRCNetwork::ToConfig() {
 		config.AddKeyValuePair("Server", m_vServers[b]->GetString());
 	}
 
+	for (const CString& sFP : m_ssTrustedFingerprints) {
+		config.AddKeyValuePair("TrustedServerFingerprint", sFP);
+	}
+
 	// Chans
 	for (unsigned int c = 0; c < m_vChans.size(); c++) {
 		CChan* pChan = m_vChans[c];
@@ -466,6 +579,8 @@ void CIRCNetwork::ClientConnected(CClient *pClient) {
 
 	size_t uIdx, uSize;
 
+	pClient->SetPlaybackActive(true);
+
 	if (m_RawBuffer.IsEmpty()) {
 		pClient->PutClient(":irc.znc.in 001 " + pClient->GetNick() + " :- Welcome to ZNC -");
 	} else {
@@ -518,19 +633,33 @@ void CIRCNetwork::ClientConnected(CClient *pClient) {
 	const vector<CChan*>& vChans = GetChans();
 	for (size_t a = 0; a < vChans.size(); a++) {
 		if ((vChans[a]->IsOn()) && (!vChans[a]->IsDetached())) {
-			vChans[a]->JoinUser(true, "", pClient);
+			vChans[a]->AttachUser(pClient);
 		}
 	}
 
-	uSize = m_QueryBuffer.Size();
+	bool bClearQuery = m_pUser->AutoClearQueryBuffer();
+	for (vector<CQuery*>::const_iterator it = m_vQueries.begin(); it != m_vQueries.end(); ++it) {
+		(*it)->SendBuffer(pClient);
+		if (bClearQuery) {
+			delete *it;
+		}
+	}
+	if (bClearQuery) {
+		m_vQueries.clear();
+	}
+
+	uSize = m_NoticeBuffer.Size();
 	for (uIdx = 0; uIdx < uSize; uIdx++) {
-		CString sLine = m_QueryBuffer.GetLine(uIdx, *pClient, msParams);
+		const CBufLine& BufLine = m_NoticeBuffer.GetBufLine(uIdx);
+		CString sLine = BufLine.GetLine(*pClient, msParams);
 		bool bContinue = false;
-		NETWORKMODULECALL(OnPrivBufferPlayLine(*pClient, sLine), m_pUser, this, NULL, &bContinue);
+		NETWORKMODULECALL(OnPrivBufferPlayLine2(*pClient, sLine, BufLine.GetTime()), m_pUser, this, NULL, &bContinue);
 		if (bContinue) continue;
 		pClient->PutClient(sLine);
 	}
-	m_QueryBuffer.Clear();
+	m_NoticeBuffer.Clear();
+
+	pClient->SetPlaybackActive(false);
 
 	// Tell them why they won't connect
 	if (!GetIRCConnectEnabled())
@@ -547,7 +676,7 @@ void CIRCNetwork::ClientDisconnected(CClient *pClient) {
 	}
 }
 
-CUser* CIRCNetwork::GetUser() {
+CUser* CIRCNetwork::GetUser() const {
 	return m_pUser;
 }
 
@@ -555,6 +684,17 @@ const CString& CIRCNetwork::GetName() const {
 	return m_sName;
 }
 
+std::vector<CClient*> CIRCNetwork::FindClients(const CString& sIdentifier) const {
+	std::vector<CClient*> vClients;
+	for (CClient* pClient : m_vClients) {
+		if (pClient->GetIdentifier().Equals(sIdentifier)) {
+			vClients.push_back(pClient);
+		}
+	}
+
+	return vClients;
+}
+
 void CIRCNetwork::SetUser(CUser *pUser) {
 	for (unsigned int a = 0; a < m_vClients.size(); a++) {
 		m_vClients[a]->PutStatus("This network is being deleted or moved to another user.");
@@ -644,6 +784,16 @@ CChan* CIRCNetwork::FindChan(CString sName) const {
 	return NULL;
 }
 
+std::vector<CChan*> CIRCNetwork::FindChans(const CString& sWild) const {
+	std::vector<CChan*> vChans;
+	vChans.reserve(m_vChans.size());
+	for (std::vector<CChan*>::const_iterator it = m_vChans.begin(); it != m_vChans.end(); ++it) {
+		if ((*it)->GetName().WildCmp(sWild))
+			vChans.push_back(*it);
+	}
+	return vChans;
+}
+
 bool CIRCNetwork::AddChan(CChan* pChan) {
 	if (!pChan) {
 		return false;
@@ -703,8 +853,11 @@ void CIRCNetwork::JoinChans() {
 			sChans.insert(pChan);
 
 			// Limit the number of joins
-			if (uJoins != 0 && --uJoins == 0)
+			if (uJoins != 0 && --uJoins == 0) {
+				// Reset the timer.
+				m_pJoinTimer->Reset();
 				break;
+			}
 		}
 	}
 
@@ -747,6 +900,12 @@ void CIRCNetwork::JoinChans(set<CChan*>& sChans) {
 }
 
 bool CIRCNetwork::JoinChan(CChan* pChan) {
+	bool bReturn = false;
+	NETWORKMODULECALL(OnJoining(*pChan), m_pUser, this, NULL, &bReturn);
+
+	if (bReturn)
+		return false;
+
 	if (m_pUser->JoinTries() != 0 && pChan->GetJoinTries() >= m_pUser->JoinTries()) {
 		PutStatus("The channel " + pChan->GetName() + " could not be joined, disabling it.");
 		pChan->Disable();
@@ -769,6 +928,64 @@ bool CIRCNetwork::IsChan(const CString& sChan) const {
 	return GetChanPrefixes().find(sChan[0]) != CString::npos;
 }
 
+// Queries
+
+const vector<CQuery*>& CIRCNetwork::GetQueries() const { return m_vQueries; }
+
+CQuery* CIRCNetwork::FindQuery(const CString& sName) const {
+	for (unsigned int a = 0; a < m_vQueries.size(); a++) {
+		CQuery* pQuery = m_vQueries[a];
+		if (sName.Equals(pQuery->GetName())) {
+			return pQuery;
+		}
+	}
+
+	return NULL;
+}
+
+std::vector<CQuery*> CIRCNetwork::FindQueries(const CString& sWild) const {
+	std::vector<CQuery*> vQueries;
+	vQueries.reserve(m_vQueries.size());
+	for (std::vector<CQuery*>::const_iterator it = m_vQueries.begin(); it != m_vQueries.end(); ++it) {
+		if ((*it)->GetName().WildCmp(sWild))
+			vQueries.push_back(*it);
+	}
+	return vQueries;
+}
+
+CQuery* CIRCNetwork::AddQuery(const CString& sName) {
+	if (sName.empty()) {
+		return NULL;
+	}
+
+	CQuery* pQuery = FindQuery(sName);
+	if (!pQuery) {
+		pQuery = new CQuery(sName, this);
+		m_vQueries.push_back(pQuery);
+
+		if (m_pUser->MaxQueryBuffers() > 0) {
+			while (m_vQueries.size() > m_pUser->MaxQueryBuffers()) {
+				delete *m_vQueries.begin();
+				m_vQueries.erase(m_vQueries.begin());
+			}
+		}
+	}
+
+	return pQuery;
+}
+
+bool CIRCNetwork::DelQuery(const CString& sName) {
+	for (vector<CQuery*>::iterator a = m_vQueries.begin(); a != m_vQueries.end(); ++a) {
+		if (sName.Equals((*a)->GetName())) {
+			delete *a;
+			m_vQueries.erase(a);
+			return true;
+		}
+	}
+
+	return false;
+}
+
 // Server list
 
 const vector<CServer*>& CIRCNetwork::GetServers() const { return m_vServers; }
@@ -976,6 +1193,7 @@ bool CIRCNetwork::Connect() {
 		return false;
 
 	if (CZNC::Get().GetServerThrottle(pServer->GetName())) {
+		// Can't connect right now, schedule retry later
 		CZNC::Get().AddNetworkToQueue(this);
 		return false;
 	}
@@ -993,6 +1211,7 @@ bool CIRCNetwork::Connect() {
 
 	CIRCSock *pIRCSock = new CIRCSock(this);
 	pIRCSock->SetPass(pServer->GetPass());
+	pIRCSock->SetSSLTrustedPeerFingerprints(m_ssTrustedFingerprints);
 
 	DEBUG("Connecting user/network [" << m_pUser->GetUserName() << "/" << m_sName << "]");
 
@@ -1021,6 +1240,14 @@ void CIRCNetwork::SetIRCSocket(CIRCSock* pIRCSock) {
 	m_pIRCSock = pIRCSock;
 }
 
+void CIRCNetwork::IRCConnected() {
+	if (m_uJoinDelay > 0) {
+		m_pJoinTimer->Delay(m_uJoinDelay);
+	} else {
+		JoinChans();
+	}
+}
+
 void CIRCNetwork::IRCDisconnected() {
 	m_pIRCSock = NULL;
 
@@ -1102,6 +1329,18 @@ const CString& CIRCNetwork::GetBindHost() const {
 	return m_sBindHost;
 }
 
+const CString& CIRCNetwork::GetEncoding() const {
+	return m_sEncoding;
+}
+
+CString CIRCNetwork::GetQuitMsg() const {
+	if (m_sQuitMsg.empty()) {
+		return m_pUser->GetQuitMsg();
+	}
+
+	return m_sQuitMsg;
+}
+
 void CIRCNetwork::SetNick(const CString& s) {
 	if (m_pUser->GetNick().Equals(s)) {
 		m_sNick = "";
@@ -1142,6 +1381,18 @@ void CIRCNetwork::SetBindHost(const CString& s) {
 	}
 }
 
+void CIRCNetwork::SetEncoding(const CString& s) {
+	m_sEncoding = s;
+}
+
+void CIRCNetwork::SetQuitMsg(const CString& s) {
+	if (m_pUser->GetQuitMsg().Equals(s)) {
+		m_sQuitMsg = "";
+	} else {
+		m_sQuitMsg = s;
+	}
+}
+
 CString CIRCNetwork::ExpandString(const CString& sStr) const {
 	CString sRet;
 	return ExpandString(sStr, sRet);
@@ -1158,3 +1409,17 @@ CString& CIRCNetwork::ExpandString(const CString& sStr, CString& sRet) const {
 
 	return m_pUser->ExpandString(sRet, sRet);
 }
+
+bool CIRCNetwork::LoadModule(const CString& sModName, const CString& sArgs, const CString& sNotice, CString& sError)
+{
+	CUtils::PrintAction(sNotice);
+	CString sModRet;
+
+	bool bModRet = GetModules().LoadModule(sModName, sArgs, CModInfo::NetworkModule, GetUser(), this, sModRet);
+
+	CUtils::PrintStatus(bModRet, sModRet);
+	if (!bModRet) {
+		sError = sModRet;
+	}
+	return bModRet;
+}
