@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2014 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2015 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,6 +19,7 @@
 #include <znc/User.h>
 #include <znc/IRCNetwork.h>
 #include <znc/znc.h>
+#include <algorithm>
 #include <sstream>
 
 using std::pair;
@@ -49,7 +50,7 @@ static CSessionManager Sessions;
 
 class CWebAuth : public CAuthBase {
 public:
-	CWebAuth(CWebSock* pWebSock, const CString& sUsername, const CString& sPassword);
+	CWebAuth(CWebSock* pWebSock, const CString& sUsername, const CString& sPassword, bool bBasic);
 	virtual ~CWebAuth() {}
 
 	void SetWebSock(CWebSock* pWebSock) { m_pWebSock = pWebSock; }
@@ -59,6 +60,7 @@ public:
 private:
 protected:
 	CWebSock*   m_pWebSock;
+	bool m_bBasic;
 };
 
 void CWebSock::FinishUserSessions(const CUser& User) {
@@ -97,13 +99,19 @@ bool CZNCTagHandler::HandleTag(CTemplate& Tmpl, const CString& sName, const CStr
 CWebSession::CWebSession(const CString& sId, const CString& sIP) : m_sId(sId), m_sIP(sIP) {
 	m_pUser = NULL;
 	Sessions.m_mIPSessions.insert(make_pair(sIP, this));
+	UpdateLastActive();
+}
+
+void CWebSession::UpdateLastActive() {
+	time(&m_tmLastActive);
 }
 
 bool CWebSession::IsAdmin() const { return IsLoggedIn() && m_pUser->IsAdmin(); }
 
-CWebAuth::CWebAuth(CWebSock* pWebSock, const CString& sUsername, const CString& sPassword)
+CWebAuth::CWebAuth(CWebSock* pWebSock, const CString& sUsername, const CString& sPassword, bool bBasic)
 	: CAuthBase(sUsername, sPassword, pWebSock) {
 	m_pWebSock = pWebSock;
+	m_bBasic = bBasic;
 }
 
 void CWebSession::ClearMessageLoops() {
@@ -147,13 +155,15 @@ void CWebSessionMap::FinishUserSessions(const CUser& User) {
 
 void CWebAuth::AcceptedLogin(CUser& User) {
 	if (m_pWebSock) {
-		CSmartPtr<CWebSession> spSession = m_pWebSock->GetSession();
+		std::shared_ptr<CWebSession> spSession = m_pWebSock->GetSession();
 
 		spSession->SetUser(&User);
 
 		m_pWebSock->SetLoggedIn(true);
 		m_pWebSock->UnPauseRead();
-		m_pWebSock->Redirect("/?cookie_check=true");
+		if (!m_bBasic) {
+			m_pWebSock->Redirect("/?cookie_check=true");
+		}
 
 		DEBUG("Successful login attempt ==> USER [" + User.GetUserName() + "] ==> SESSION [" + spSession->GetId() + "]");
 	}
@@ -161,7 +171,7 @@ void CWebAuth::AcceptedLogin(CUser& User) {
 
 void CWebAuth::RefusedLogin(const CString& sReason) {
 	if (m_pWebSock) {
-		CSmartPtr<CWebSession> spSession = m_pWebSock->GetSession();
+		std::shared_ptr<CWebSession> spSession = m_pWebSock->GetSession();
 
 		spSession->AddError("Invalid login!");
 		spSession->SetUser(NULL);
@@ -179,14 +189,14 @@ void CWebAuth::Invalidate() {
 	m_pWebSock = NULL;
 }
 
-CWebSock::CWebSock() : CHTTPSock(NULL) {
+CWebSock::CWebSock(const CString& sURIPrefix) : CHTTPSock(NULL, sURIPrefix) {
 	m_bPathsSet = false;
 
-	m_Template.AddTagHandler(new CZNCTagHandler(*this));
+	m_Template.AddTagHandler(std::make_shared<CZNCTagHandler>(*this));
 }
 
 CWebSock::~CWebSock() {
-	if (!m_spAuth.IsNull()) {
+	if (m_spAuth) {
 		m_spAuth->Invalidate();
 	}
 
@@ -319,6 +329,7 @@ void CWebSock::SetVars() {
 	m_Template["Version"] = CZNC::GetVersion();
 	m_Template["SkinName"] = GetSkinName();
 	m_Template["_CSRF_Check"] = GetCSRFCheck();
+	m_Template["URIPrefix"] = GetURIPrefix();
 
 	if (GetSession()->IsAdmin()) {
 		m_Template["IsAdmin"] = "true";
@@ -467,6 +478,7 @@ CWebSock::EPageReqResult CWebSock::PrintStaticFile(const CString& sPath, CString
 
 CWebSock::EPageReqResult CWebSock::PrintTemplate(const CString& sPageName, CString& sPageRet, CModule* pModule) {
 	SetVars();
+
 	m_Template["PageName"] = sPageName;
 
 	if (pModule) {
@@ -620,7 +632,7 @@ CWebSock::EPageReqResult CWebSock::OnPageRequestInternal(const CString& sURI, CS
 		if (GetParam("submitted").ToBool()) {
 			m_sUser = GetParam("user");
 			m_sPass = GetParam("pass");
-			m_bLoggedIn = OnLogin(m_sUser, m_sPass);
+			m_bLoggedIn = OnLogin(m_sUser, m_sPass, false);
 
 			// AcceptedLogin()/RefusedLogin() will call Redirect()
 			return PAGE_DEFERRED;
@@ -807,24 +819,31 @@ void CWebSock::PrintErrorPage(const CString& sMessage) {
 	m_Template["Error"] = sMessage;
 }
 
-CSmartPtr<CWebSession> CWebSock::GetSession() {
-	if (!m_spSession.IsNull()) {
+static inline bool compareLastActive(const std::pair<const CString, CWebSession *> &first, const std::pair<const CString, CWebSession *> &second) {
+	return first.second->GetLastActive() < second.second->GetLastActive();
+}
+
+std::shared_ptr<CWebSession> CWebSock::GetSession() {
+	if (m_spSession) {
 		return m_spSession;
 	}
 
 	const CString sCookieSessionId = GetRequestCookie("SessionId");
-	CSmartPtr<CWebSession> *pSession = Sessions.m_mspSessions.GetItem(sCookieSessionId);
+	std::shared_ptr<CWebSession> *pSession = Sessions.m_mspSessions.GetItem(sCookieSessionId);
 
 	if (pSession != NULL) {
 		// Refresh the timeout
 		Sessions.m_mspSessions.AddItem((*pSession)->GetId(), *pSession);
+		(*pSession)->UpdateLastActive();
 		m_spSession = *pSession;
-		DEBUG("Found existing session from cookie: [" + sCookieSessionId + "] IsLoggedIn(" + CString((*pSession)->IsLoggedIn() ? "true" : "false") + ")");
+		DEBUG("Found existing session from cookie: [" + sCookieSessionId + "] IsLoggedIn(" + CString((*pSession)->IsLoggedIn() ? "true, " + ((*pSession)->GetUser()->GetUserName()) : "false") + ")");
 		return *pSession;
 	}
 
 	if (Sessions.m_mIPSessions.count(GetRemoteIP()) > m_uiMaxSessions) {
-		mIPSessionsIterator it = Sessions.m_mIPSessions.find(GetRemoteIP());
+		pair<mIPSessionsIterator, mIPSessionsIterator> p =
+			Sessions.m_mIPSessions.equal_range(GetRemoteIP());
+		mIPSessionsIterator it = std::min_element(p.first, p.second, compareLastActive);
 		DEBUG("Remote IP:   " << GetRemoteIP() << "; discarding session [" << it->second->GetId() << "]");
 		Sessions.m_mspSessions.RemItem(it->second->GetId());
 	}
@@ -840,7 +859,7 @@ CSmartPtr<CWebSession> CWebSock::GetSession() {
 		DEBUG("Auto generated session: [" + sSessionID + "]");
 	} while (Sessions.m_mspSessions.HasItem(sSessionID));
 
-	CSmartPtr<CWebSession> spSession(new CWebSession(sSessionID, GetRemoteIP()));
+	std::shared_ptr<CWebSession> spSession(new CWebSession(sSessionID, GetRemoteIP()));
 	Sessions.m_mspSessions.AddItem(spSession->GetId(), spSession);
 
 	m_spSession = spSession;
@@ -849,13 +868,13 @@ CSmartPtr<CWebSession> CWebSock::GetSession() {
 }
 
 CString CWebSock::GetCSRFCheck() {
-	CSmartPtr<CWebSession> pSession = GetSession();
+	std::shared_ptr<CWebSession> pSession = GetSession();
 	return pSession->GetId().MD5();
 }
 
-bool CWebSock::OnLogin(const CString& sUser, const CString& sPass) {
-	DEBUG("=================== CWebSock::OnLogin()");
-	m_spAuth = new CWebAuth(this, sUser, sPass);
+bool CWebSock::OnLogin(const CString& sUser, const CString& sPass, bool bBasic) {
+	DEBUG("=================== CWebSock::OnLogin(), basic auth? " << std::boolalpha << bBasic);
+	m_spAuth = std::make_shared<CWebAuth>(this, sUser, sPass, bBasic);
 
 	// Some authentication module could need some time, block this socket
 	// until then. CWebAuth will UnPauseRead().
@@ -874,7 +893,7 @@ Csock* CWebSock::GetSockObj(const CString& sHost, unsigned short uPort) {
 }
 
 CString CWebSock::GetSkinName() {
-	CSmartPtr<CWebSession> spSession = GetSession();
+	std::shared_ptr<CWebSession> spSession = GetSession();
 
 	if (spSession->IsLoggedIn() && !spSession->GetUser()->GetSkinName().empty()) {
 		return spSession->GetUser()->GetSkinName();
