@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2014 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2015 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -20,6 +20,7 @@
 #include <znc/User.h>
 #include <znc/IRCNetwork.h>
 #include <znc/Config.h>
+#include <tuple>
 
 using std::endl;
 using std::cout;
@@ -27,6 +28,8 @@ using std::map;
 using std::set;
 using std::vector;
 using std::list;
+using std::tuple;
+using std::make_tuple;
 
 static inline CString FormatBindError() {
 	CString sError = (errno == 0 ? CString("unknown error, check the host name") : CString(strerror(errno)));
@@ -52,6 +55,9 @@ CZNC::CZNC() {
 	m_sConnectThrottle.SetTTL(30000);
 	m_pLockFile = NULL;
 	m_bProtectWebSessions = true;
+	m_bHideVersion = false;
+	m_uDisabledSSLProtocols = Csock::EDP_SSL;
+	m_sSSLProtocols = "";
 }
 
 CZNC::~CZNC() {
@@ -87,28 +93,22 @@ CZNC::~CZNC() {
 }
 
 CString CZNC::GetVersion() {
-	char szBuf[128];
-
-	snprintf(szBuf, sizeof(szBuf), "%1.1f%s", VERSION, ZNC_VERSION_EXTRA);
-	// If snprintf overflows (which I doubt), we want to be on the safe side
-	szBuf[sizeof(szBuf) - 1] = '\0';
-
-	return szBuf;
+	return CString(VERSION_STR) + CString(ZNC_VERSION_EXTRA);
 }
 
 CString CZNC::GetTag(bool bIncludeVersion, bool bHTML) {
+	if (!Get().m_bHideVersion) {
+		bIncludeVersion = true;
+	}
 	CString sAddress = bHTML ? "<a href=\"http://znc.in\">http://znc.in</a>" : "http://znc.in";
 
 	if (!bIncludeVersion) {
 		return "ZNC - " + sAddress;
 	}
 
-	char szBuf[128];
-	snprintf(szBuf, sizeof(szBuf), "ZNC %1.1f%s - ", VERSION, ZNC_VERSION_EXTRA);
-	// If snprintf overflows (which I doubt), we want to be on the safe side
-	szBuf[sizeof(szBuf) - 1] = '\0';
+	CString sVersion = GetVersion();
 
-	return szBuf + sAddress;
+	return "ZNC - " + sVersion + " - " + sAddress;
 }
 
 CString CZNC::GetCompileOptionsString() {
@@ -130,6 +130,12 @@ CString CZNC::GetCompileOptionsString() {
 		"threads"
 #else
 		"blocking"
+#endif
+		", charset: "
+#ifdef HAVE_ICU
+		"yes"
+#else
+		"no"
 #endif
 	;
 }
@@ -178,7 +184,8 @@ void CZNC::Loop() {
 	while (true) {
 		CString sError;
 
-		switch (GetConfigState()) {
+		ConfigState eState = GetConfigState();
+		switch (eState) {
 		case ECONFIG_NEED_REHASH:
 			SetConfigState(ECONFIG_NOTHING);
 
@@ -190,12 +197,13 @@ void CZNC::Loop() {
 			}
 			break;
 		case ECONFIG_NEED_WRITE:
+		case ECONFIG_NEED_VERBOSE_WRITE:
 			SetConfigState(ECONFIG_NOTHING);
 
-			if (WriteConfig()) {
-				Broadcast("Writing the config succeeded", true);
-			} else {
+			if (!WriteConfig()) {
 				Broadcast("Writing the config file failed", true);
+			} else if (eState == ECONFIG_NEED_VERBOSE_WRITE) {
+				Broadcast("Writing the config succeeded", true);
 			}
 			break;
 		case ECONFIG_NOTHING:
@@ -437,13 +445,15 @@ bool CZNC::WriteConfig() {
 	config.AddKeyValuePair("MaxBufferSize", CString(m_uiMaxBufferSize));
 	config.AddKeyValuePair("SSLCertFile", CString(m_sSSLCertFile));
 	config.AddKeyValuePair("ProtectWebSessions", CString(m_bProtectWebSessions));
-	config.AddKeyValuePair("Version", CString(VERSION, 1));
+	config.AddKeyValuePair("HideVersion", CString(m_bHideVersion));
+	config.AddKeyValuePair("Version", CString(VERSION_STR));
 
 	for (size_t l = 0; l < m_vpListeners.size(); l++) {
 		CListener* pListener = m_vpListeners[l];
 		CConfig listenerConfig;
 
 		listenerConfig.AddKeyValuePair("Host", pListener->GetBindHost());
+		listenerConfig.AddKeyValuePair("URIPrefix", pListener->GetURIPrefix() + "/");
 		listenerConfig.AddKeyValuePair("Port", CString(pListener->GetPort()));
 
 		listenerConfig.AddKeyValuePair("IPv4", CString(pListener->GetAddrType() != ADDR_IPV6ONLY));
@@ -472,6 +482,14 @@ bool CZNC::WriteConfig() {
 		config.AddKeyValuePair("StatusPrefix", m_sStatusPrefix.FirstLine());
 	}
 
+	if (!m_sSSLCiphers.empty()) {
+		config.AddKeyValuePair("SSLCiphers", CString(m_sSSLCiphers));
+	}
+
+	if (!m_sSSLProtocols.empty()) {
+		config.AddKeyValuePair("SSLProtocols", m_sSSLProtocols);
+	}
+
 	for (unsigned int m = 0; m < m_vsMotd.size(); m++) {
 		config.AddKeyValuePair("Motd", m_vsMotd[m].FirstLine());
 	}
@@ -480,6 +498,10 @@ bool CZNC::WriteConfig() {
 		config.AddKeyValuePair("BindHost", m_vsBindHosts[v].FirstLine());
 	}
 
+	for (unsigned int v = 0; v < m_vsTrustedProxies.size(); v++) {
+		config.AddKeyValuePair("TrustedProxy", m_vsTrustedProxies[v].FirstLine());
+	}
+
 	CModules& Mods = GetModules();
 
 	for (unsigned int a = 0; a < Mods.size(); a++) {
@@ -552,13 +574,16 @@ bool CZNC::WriteNewConfig(const CString& sConfigFile) {
 	VCString vsLines;
 
 	vsLines.push_back(MakeConfigHeader());
-	vsLines.push_back("Version = " + CString(VERSION, 1));
+	vsLines.push_back("Version = " + CString(VERSION_STR));
 
 	m_sConfigFile = ExpandConfigPath(sConfigFile);
-	CUtils::PrintMessage("Building new config");
+
+	if (CFile::Exists(m_sConfigFile)) {
+		CUtils::PrintStatus(false, "WARNING: config [" + m_sConfigFile + "] already exists.");
+	}
 
 	CUtils::PrintMessage("");
-	CUtils::PrintMessage("First let's start with some global settings...");
+	CUtils::PrintMessage("-- Global settings --");
 	CUtils::PrintMessage("");
 
 	// Listen
@@ -568,6 +593,7 @@ bool CZNC::WriteNewConfig(const CString& sConfigFile) {
 	bool b6 = false;
 #endif
 	CString sListenHost;
+	CString sURIPrefix;
 	bool bListenSSL = false;
 	unsigned int uListenPort = 0;
 	bool bSuccess;
@@ -575,28 +601,32 @@ bool CZNC::WriteNewConfig(const CString& sConfigFile) {
 	do {
 		bSuccess = true;
 		while (true) {
-			if (!CUtils::GetNumInput("What port would you like ZNC to listen on?", uListenPort, 1025, 65535)) {
+			if (!CUtils::GetNumInput("Listen on port", uListenPort, 1025, 65534)) {
 				continue;
 			}
-			if (uListenPort == 6667 && !CUtils::GetBoolInput("Warning: Some web browsers reject port 6667. If you intend to use ZNC's web interface, you might want to use another port. Proceed with port 6667 anyway?", true)) {
-				continue;
+			if (uListenPort == 6667) {
+				CUtils::PrintStatus(false, "WARNING: Some web browsers reject port 6667. If you intend to");
+				CUtils::PrintStatus(false, "use ZNC's web interface, you might want to use another port.");
+				if (!CUtils::GetBoolInput("Proceed with port 6667 anyway?", true)) {
+					continue;
+				}
 			}
 			break;
 		}
 
 
 #ifdef HAVE_LIBSSL
-		bListenSSL = CUtils::GetBoolInput("Would you like ZNC to listen using SSL?", bListenSSL);
+		bListenSSL = CUtils::GetBoolInput("Listen using SSL", bListenSSL);
 #endif
 
 #ifdef HAVE_IPV6
-		b6 = CUtils::GetBoolInput("Would you like ZNC to listen using both IPv4 and IPv6?", b6);
+		b6 = CUtils::GetBoolInput("Listen using both IPv4 and IPv6", b6);
 #endif
 
 		// Don't ask for listen host, it may be configured later if needed.
 
 		CUtils::PrintAction("Verifying the listener");
-		CListener* pListener = new CListener((unsigned short int)uListenPort, sListenHost, bListenSSL,
+		CListener* pListener = new CListener((unsigned short int)uListenPort, sListenHost, sURIPrefix, bListenSSL,
 				b6 ? ADDR_ALL : ADDR_IPV4ONLY, CListener::ACCEPT_ALL);
 		if (!pListener->Listen()) {
 			CUtils::PrintStatus(false, FormatBindError());
@@ -626,265 +656,121 @@ bool CZNC::WriteNewConfig(const CString& sConfigFile) {
 	// !Listen
 
 	set<CModInfo> ssGlobalMods;
-	GetModules().GetAvailableMods(ssGlobalMods, CModInfo::GlobalModule);
-	size_t uNrOtherGlobalMods = FilterUncommonModules(ssGlobalMods);
-
-	if (!ssGlobalMods.empty()) {
-		CUtils::PrintMessage("");
-		CUtils::PrintMessage("-- Global Modules --");
-		CUtils::PrintMessage("");
-
-		CTable Table;
-		Table.AddColumn("Name");
-		Table.AddColumn("Description");
-		set<CModInfo>::iterator it;
-
-		for (it = ssGlobalMods.begin(); it != ssGlobalMods.end(); ++it) {
-			const CModInfo& Info = *it;
-			Table.AddRow();
-			Table.SetCell("Name", Info.GetName());
-			Table.SetCell("Description", Info.GetDescription().Ellipsize(128));
-		}
-
-		unsigned int uTableIdx = 0; CString sLine;
-		while (Table.GetLine(uTableIdx++, sLine)) {
-			CUtils::PrintMessage(sLine);
-		}
-
-		if (uNrOtherGlobalMods > 0) {
-			CUtils::PrintMessage("And " + CString(uNrOtherGlobalMods) + " other (uncommon) modules. You can enable those later.");
-		}
-
-		CUtils::PrintMessage("");
-
-		for (it = ssGlobalMods.begin(); it != ssGlobalMods.end(); ++it) {
-			const CModInfo& Info = *it;
-			CString sName = Info.GetName();
-
-			if (CDebug::StdoutIsTTY()) {
-				if (CUtils::GetBoolInput("Load global module <\033[1m" + sName + "\033[22m>?", false))
-					vsLines.push_back("LoadModule = " + sName);
-			} else {
-				if (CUtils::GetBoolInput("Load global module <" + sName + ">?", false))
-					vsLines.push_back("LoadModule = " + sName);
-			}
-		}
+	GetModules().GetDefaultMods(ssGlobalMods, CModInfo::GlobalModule);
+	vector<CString> vsGlobalModNames;
+	for (set<CModInfo>::const_iterator it = ssGlobalMods.begin(); it != ssGlobalMods.end(); ++it) {
+		vsGlobalModNames.push_back(it->GetName());
+		vsLines.push_back("LoadModule = " + it->GetName());
 	}
+	CUtils::PrintMessage("Enabled global modules [" + CString(", ").Join(vsGlobalModNames.begin(), vsGlobalModNames.end()) + "]");
 
 	// User
 	CUtils::PrintMessage("");
-	CUtils::PrintMessage("Now we need to set up a user...");
+	CUtils::PrintMessage("-- Admin user settings --");
 	CUtils::PrintMessage("");
 
-	bool bFirstUser = true;
-
+	vsLines.push_back("");
+	CString sNick;
 	do {
+		CUtils::GetInput("Username", sUser, "", "alphanumeric");
+	} while (!CUser::IsValidUserName(sUser));
+
+	vsLines.push_back("<User " + sUser + ">");
+	CString sSalt;
+	sAnswer = CUtils::GetSaltedHashPass(sSalt);
+	vsLines.push_back("\tPass       = " + CUtils::sDefaultHash + "#" + sAnswer + "#" + sSalt + "#");
+
+	vsLines.push_back("\tAdmin      = true");
+
+	CUtils::GetInput("Nick", sNick, CUser::MakeCleanUserName(sUser));
+	vsLines.push_back("\tNick       = " + sNick);
+	CUtils::GetInput("Alternate nick", sAnswer, sNick + "_");
+	if (!sAnswer.empty()) {
+		vsLines.push_back("\tAltNick    = " + sAnswer);
+	}
+	CUtils::GetInput("Ident", sAnswer, sUser);
+	vsLines.push_back("\tIdent      = " + sAnswer);
+	CUtils::GetInput("Real name", sAnswer, "Got ZNC?");
+	vsLines.push_back("\tRealName   = " + sAnswer);
+	CUtils::GetInput("Bind host", sAnswer, "", "optional");
+	if (!sAnswer.empty()) {
+		vsLines.push_back("\tBindHost   = " + sAnswer);
+	}
+
+	set<CModInfo> ssUserMods;
+	GetModules().GetDefaultMods(ssUserMods, CModInfo::UserModule);
+	vector<CString> vsUserModNames;
+	for (set<CModInfo>::const_iterator it = ssUserMods.begin(); it != ssUserMods.end(); ++it) {
+		vsUserModNames.push_back(it->GetName());
+		vsLines.push_back("\tLoadModule = " + it->GetName());
+	}
+	CUtils::PrintMessage("Enabled user modules [" + CString(", ").Join(vsUserModNames.begin(), vsUserModNames.end()) + "]");
+
+	CUtils::PrintMessage("");
+	if (CUtils::GetBoolInput("Set up a network?", true)) {
 		vsLines.push_back("");
-		CString sNick;
+
+		CUtils::PrintMessage("");
+		CUtils::PrintMessage("-- Network settings --");
+		CUtils::PrintMessage("");
+
 		do {
-			CUtils::GetInput("Username", sUser, "", "AlphaNumeric");
-		} while (!CUser::IsValidUserName(sUser));
+			CUtils::GetInput("Name", sNetwork, "freenode");
+		} while (!CIRCNetwork::IsValidNetwork(sNetwork));
 
-		vsLines.push_back("<User " + sUser + ">");
-		CString sSalt;
-		sAnswer = CUtils::GetSaltedHashPass(sSalt);
-		vsLines.push_back("\tPass       = " + CUtils::sDefaultHash + "#" + sAnswer + "#" + sSalt + "#");
+		vsLines.push_back("\t<Network " + sNetwork + ">");
 
-		if (CUtils::GetBoolInput("Would you like this user to be an admin?", bFirstUser)) {
-			vsLines.push_back("\tAdmin      = true");
-		} else {
-			vsLines.push_back("\tAdmin      = false");
+		set<CModInfo> ssNetworkMods;
+		GetModules().GetDefaultMods(ssNetworkMods, CModInfo::NetworkModule);
+		vector<CString> vsNetworkModNames;
+		for (set<CModInfo>::const_iterator it = ssNetworkMods.begin(); it != ssNetworkMods.end(); ++it) {
+			vsNetworkModNames.push_back(it->GetName());
+			vsLines.push_back("\t\tLoadModule = " + it->GetName());
 		}
 
-		CUtils::GetInput("Nick", sNick, CUser::MakeCleanUserName(sUser));
-		vsLines.push_back("\tNick       = " + sNick);
-		CUtils::GetInput("Alt Nick", sAnswer, sNick + "_");
-		if (!sAnswer.empty()) {
-			vsLines.push_back("\tAltNick    = " + sAnswer);
-		}
-		CUtils::GetInput("Ident", sAnswer, sNick);
-		vsLines.push_back("\tIdent      = " + sAnswer);
-		CUtils::GetInput("Real Name", sAnswer, "Got ZNC?");
-		vsLines.push_back("\tRealName   = " + sAnswer);
-		CUtils::GetInput("Bind Host", sAnswer, "", "optional");
-		if (!sAnswer.empty()) {
-			vsLines.push_back("\tBindHost   = " + sAnswer);
-		}
-		// todo: Possibly add motd
-
-		unsigned int uBufferCount = 0;
-
-		CUtils::GetNumInput("Number of lines to buffer per channel", uBufferCount, 0, ~0, 50);
-		if (uBufferCount) {
-			vsLines.push_back("\tBuffer     = " + CString(uBufferCount));
-		}
-		if (CUtils::GetBoolInput("Would you like to clear channel buffers after replay?", true)) {
-			vsLines.push_back("\tAutoClearChanBuffer = true");
-		} else {
-			vsLines.push_back("\tAutoClearChanBuffer = false");
-		}
-
-		CUtils::GetInput("Default channel modes", sAnswer, "+stn");
-		if (!sAnswer.empty()) {
-			vsLines.push_back("\tChanModes  = " + sAnswer);
-		}
-
-		set<CModInfo> ssUserMods;
-		GetModules().GetAvailableMods(ssUserMods);
-		size_t uNrOtherUserMods = FilterUncommonModules(ssUserMods);
-
-		if (!ssUserMods.empty()) {
-			vsLines.push_back("");
-			CUtils::PrintMessage("");
-			CUtils::PrintMessage("-- User Modules --");
-			CUtils::PrintMessage("");
-
-			CTable Table;
-			Table.AddColumn("Name");
-			Table.AddColumn("Description");
-			set<CModInfo>::iterator it;
-
-			for (it = ssUserMods.begin(); it != ssUserMods.end(); ++it) {
-				const CModInfo& Info = *it;
-				Table.AddRow();
-				Table.SetCell("Name", Info.GetName());
-				Table.SetCell("Description", Info.GetDescription().Ellipsize(128));
-			}
-
-			unsigned int uTableIdx = 0; CString sLine;
-			while (Table.GetLine(uTableIdx++, sLine)) {
-				CUtils::PrintMessage(sLine);
-			}
-
-			if (uNrOtherUserMods > 0) {
-				CUtils::PrintMessage("And " + CString(uNrOtherUserMods) + " other (uncommon) modules. You can enable those later.");
-			}
-
-			CUtils::PrintMessage("");
-
-			for (it = ssUserMods.begin(); it != ssUserMods.end(); ++it) {
-				const CModInfo& Info = *it;
-				CString sName = Info.GetName();
-
-				if (CDebug::StdoutIsTTY()) {
-					if (CUtils::GetBoolInput("Load module <\033[1m" + sName + "\033[22m>?", false))
-						vsLines.push_back("\tLoadModule = " + sName);
-				} else {
-					if (CUtils::GetBoolInput("Load module <" + sName + ">?", false))
-						vsLines.push_back("\tLoadModule = " + sName);
-				}
-			}
-		}
-
-		CUtils::PrintMessage("");
-		CString sAAnother = "a";
-		while (CUtils::GetBoolInput("Would you like to set up " + sAAnother + " network?", false)) {
-			sAAnother = "another";
-			vsLines.push_back("");
-
-			do {
-				CUtils::GetInput("Network", sNetwork, "", "e.g. `freenode' or `efnet'");
-			} while (!CIRCNetwork::IsValidNetwork(sNetwork));
-
-			vsLines.push_back("\t<Network " + sNetwork + ">");
-
-			set<CModInfo> ssNetworkMods;
-			GetModules().GetAvailableMods(ssNetworkMods, CModInfo::NetworkModule);
-			size_t uNrOtherNetworkMods = FilterUncommonModules(ssNetworkMods);
-
-			if (!ssNetworkMods.empty()) {
-				CUtils::PrintMessage("");
-				CUtils::PrintMessage("-- Network Modules --");
-				CUtils::PrintMessage("");
-
-				CTable Table;
-				Table.AddColumn("Name");
-				Table.AddColumn("Description");
-				set<CModInfo>::iterator it;
-
-				for (it = ssNetworkMods.begin(); it != ssNetworkMods.end(); ++it) {
-					const CModInfo& Info = *it;
-					Table.AddRow();
-					Table.SetCell("Name", Info.GetName());
-					Table.SetCell("Description", Info.GetDescription().Ellipsize(128));
-				}
-
-				unsigned int uTableIdx = 0; CString sLine;
-				while (Table.GetLine(uTableIdx++, sLine)) {
-					CUtils::PrintMessage(sLine);
-				}
-
-				if (uNrOtherNetworkMods > 0) {
-					CUtils::PrintMessage("And " + CString(uNrOtherNetworkMods) + " other (uncommon) modules. You can enable those later.");
-				}
-
-				CUtils::PrintMessage("");
-
-				for (it = ssNetworkMods.begin(); it != ssNetworkMods.end(); ++it) {
-					const CModInfo& Info = *it;
-					CString sName = Info.GetName();
-
-					if (CDebug::StdoutIsTTY()) {
-						if (CUtils::GetBoolInput("Load module <\033[1m" + sName + "\033[22m>?", false))
-							vsLines.push_back("\t\tLoadModule = " + sName);
-					} else {
-						if (CUtils::GetBoolInput("Load module <" + sName + ">?", false))
-							vsLines.push_back("\t\tLoadModule = " + sName);
-					}
-				}
-			}
-
-			vsLines.push_back("");
-			CUtils::PrintMessage("");
-			CUtils::PrintMessage("-- IRC Servers --");
-			CUtils::PrintMessage("Only add servers from the same IRC network.");
-			CUtils::PrintMessage("If a server from the list can't be reached, another server will be used.");
-			CUtils::PrintMessage("");
-
-			do {
-				CString sHost, sPass;
-				bool bSSL = false;
-				unsigned int uServerPort = 0;
-
-				while (!CUtils::GetInput("IRC server", sHost, "", "host only") || !CServer::IsValidHostName(sHost)) ;
-				while (!CUtils::GetNumInput("[" + sHost + "] Port", uServerPort, 1, 65535, 6667)) ;
-				CUtils::GetInput("[" + sHost + "] Password (probably empty)", sPass);
+		CString sHost, sPass, sHint;
+		bool bSSL = false;
+		unsigned int uServerPort = 0;
 
+		if (sNetwork.Equals("freenode")) {
+			sHost = "chat.freenode.net";
 #ifdef HAVE_LIBSSL
-				bSSL = CUtils::GetBoolInput("Does this server use SSL?", false);
+			bSSL = true;
 #endif
-
-				vsLines.push_back("\t\tServer     = " + sHost + ((bSSL) ? " +" : " ") + CString(uServerPort) + " " + sPass);
-
-				CUtils::PrintMessage("");
-			} while (CUtils::GetBoolInput("Would you like to add another server for this IRC network?", false));
-
-			vsLines.push_back("");
-			CUtils::PrintMessage("");
-			CUtils::PrintMessage("-- Channels --");
-			CUtils::PrintMessage("");
-
-			CString sArg = "a";
-			CString sPost = " for ZNC to automatically join?";
-			bool bDefault = true;
-
-			while (CUtils::GetBoolInput("Would you like to add " + sArg + " channel" + sPost, bDefault)) {
-				while (!CUtils::GetInput("Channel name", sAnswer)) ;
-				vsLines.push_back("\t\t<Chan " + sAnswer + ">");
-				vsLines.push_back("\t\t</Chan>");
-				sArg = "another";
-				sPost = "?";
-				bDefault = false;
-			}
-
-			vsLines.push_back("\t</Network>");
+		} else {
+			sHint = "host only";
 		}
 
-		vsLines.push_back("</User>");
+		while (!CUtils::GetInput("Server host", sHost, sHost, sHint) || !CServer::IsValidHostName(sHost));
+#ifdef HAVE_LIBSSL
+		bSSL = CUtils::GetBoolInput("Server uses SSL?", bSSL);
+#endif
+		while (!CUtils::GetNumInput("Server port", uServerPort, 1, 65535, bSSL ? 6697 : 6667));
+		CUtils::GetInput("Server password (probably empty)", sPass);
 
-		CUtils::PrintMessage("");
-		bFirstUser = false;
-	} while (CUtils::GetBoolInput("Would you like to set up another user?", false));
+		vsLines.push_back("\t\tServer     = " + sHost + ((bSSL) ? " +" : " ") + CString(uServerPort) + " " + sPass);
+
+		CString sChans;
+		if (CUtils::GetInput("Initial channels", sChans)) {
+			vsLines.push_back("");
+			VCString vsChans;
+			sChans.Replace(",", " ");
+			sChans.Replace(";", " ");
+			sChans.Split(" ", vsChans, false, "", "", true, true);
+			for (const CString& sChan : vsChans) {
+				vsLines.push_back("\t\t<Chan " + sChan + ">");
+				vsLines.push_back("\t\t</Chan>");
+			}
+		}
+
+		CUtils::PrintMessage("Enabled network modules [" + CString(", ").Join(vsNetworkModNames.begin(), vsNetworkModNames.end()) + "]");
+
+		vsLines.push_back("\t</Network>");
+	}
+
+	vsLines.push_back("</User>");
+
+	CUtils::PrintMessage("");
 	// !User
 
 	CFile File;
@@ -900,7 +786,7 @@ bool CZNC::WriteNewConfig(const CString& sConfigFile) {
 			} else {
 				File.Close();
 				CUtils::PrintStatus(false, "This config already exists.");
-				if (CUtils::GetBoolInput("Would you like to overwrite it?", false))
+				if (CUtils::GetBoolInput("Are you sure you want to overwrite it?", false))
 					CUtils::PrintAction("Overwriting config [" + m_sConfigFile + "]");
 				else
 					bFileOK = false;
@@ -917,7 +803,7 @@ bool CZNC::WriteNewConfig(const CString& sConfigFile) {
 			}
 		}
 		if (!bFileOK) {
-			CUtils::GetInput("Please specify an alternate location (or \"stdout\" for displaying the config)", m_sConfigFile, m_sConfigFile);
+			while (!CUtils::GetInput("Please specify an alternate location", m_sConfigFile, "",  "or \"stdout\" for displaying the config"));
 			if (m_sConfigFile.Equals("stdout"))
 				bFileOK = true;
 			else
@@ -969,7 +855,8 @@ bool CZNC::WriteNewConfig(const CString& sConfigFile) {
 	CUtils::PrintMessage("");
 	CUtils::PrintMessage("Try something like this in your IRC client...", true);
 	CUtils::PrintMessage("/server <znc_server_ip> " + sSSL + CString(uListenPort) + " " + sUser + ":<pass>", true);
-	CUtils::PrintMessage("And this in your browser...", true);
+	CUtils::PrintMessage("");
+	CUtils::PrintMessage("To manage settings, users and networks, point your web browser to", true);
 	CUtils::PrintMessage(sProtocol + "://<znc_server_ip>:" + CString(uListenPort) + "/", true);
 	CUtils::PrintMessage("");
 
@@ -977,26 +864,6 @@ bool CZNC::WriteNewConfig(const CString& sConfigFile) {
 	return bFileOpen && CUtils::GetBoolInput("Launch ZNC now?", true);
 }
 
-size_t CZNC::FilterUncommonModules(set<CModInfo>& ssModules) {
-	const char* ns[] = { "webadmin", "controlpanel",
-		"chansaver", "keepnick", "simple_away", "partyline",
-		"kickrejoin", "nickserv", "perform" };
-	const set<CString> ssNames(ns, ns + sizeof(ns) / sizeof(ns[0]));
-
-	size_t uNrRemoved = 0;
-	for(set<CModInfo>::iterator it = ssModules.begin(); it != ssModules.end(); ) {
-		if(ssNames.count(it->GetName()) > 0) {
-			++it;
-		} else {
-			set<CModInfo>::iterator it2 = it++;
-			ssModules.erase(it2);
-			uNrRemoved++;
-		}
-	}
-
-	return uNrRemoved;
-}
-
 void CZNC::BackupConfigOnce(const CString& sSuffix) {
 	static bool didBackup = false;
 	if (didBackup)
@@ -1012,13 +879,11 @@ void CZNC::BackupConfigOnce(const CString& sSuffix) {
 		CUtils::PrintStatus(false, strerror(errno));
 }
 
-bool CZNC::ParseConfig(const CString& sConfig)
+bool CZNC::ParseConfig(const CString& sConfig, CString& sError)
 {
-	CString s;
-
 	m_sConfigFile = ExpandConfigPath(sConfig, false);
 
-	return DoRehash(s);
+	return DoRehash(sError);
 }
 
 bool CZNC::RehashConfig(CString& sError)
@@ -1099,18 +964,21 @@ bool CZNC::DoRehash(CString& sError)
 
 	CString sSavedVersion;
 	config.FindStringEntry("version", sSavedVersion);
-	double fSavedVersion = sSavedVersion.ToDouble();
-	if (fSavedVersion < VERSION - 0.000001) {
+	tuple<unsigned int, unsigned int> tSavedVersion = make_tuple(sSavedVersion.Token(0, false, ".").ToUInt(),
+																 sSavedVersion.Token(1, false, ".").ToUInt());
+	tuple<unsigned int, unsigned int> tCurrentVersion = make_tuple(VERSION_MAJOR, VERSION_MINOR);
+	if (tSavedVersion < tCurrentVersion) {
 		if (sSavedVersion.empty()) {
 			sSavedVersion = "< 0.203";
 		}
 		CUtils::PrintMessage("Found old config from ZNC " + sSavedVersion + ". Saving a backup of it.");
-		BackupConfigOnce("pre-" + CString(VERSION, 1));
-	} else if (fSavedVersion > VERSION + 0.000001) {
+		BackupConfigOnce("pre-" + CString(VERSION_STR));
+	} else if (tSavedVersion > tCurrentVersion) {
 		CUtils::PrintError("Config was saved from ZNC " + sSavedVersion + ". It may or may not work with current ZNC " + GetVersion());
 	}
 
 	m_vsBindHosts.clear();
+	m_vsTrustedProxies.clear();
 	m_vsMotd.clear();
 
 	// Delete all listeners
@@ -1128,7 +996,7 @@ bool CZNC::DoRehash(CString& sError)
 		CString sModName = vit->Token(0);
 		CString sArgs = vit->Token(1, true);
 
-		if (sModName == "saslauth" && fSavedVersion < 0.207 + 0.000001) {
+		if (sModName == "saslauth" && tSavedVersion < make_tuple(0, 207)) {
 			// XXX compatibility crap, added in 0.207
 			CUtils::PrintMessage("saslauth module was renamed to cyrusauth. Loading cyrusauth instead.");
 			sModName = "cyrusauth";
@@ -1204,6 +1072,12 @@ bool CZNC::DoRehash(CString& sError)
 	for (vit = vsList.begin(); vit != vsList.end(); ++vit) {
 		AddBindHost(*vit);
 	}
+
+	config.FindStringVector("trustedproxy", vsList);
+	for (vit = vsList.begin(); vit != vsList.end(); ++vit) {
+		AddTrustedProxy(*vit);
+	}
+
 	config.FindStringVector("vhost", vsList);
 	for (vit = vsList.begin(); vit != vsList.end(); ++vit) {
 		AddBindHost(*vit);
@@ -1216,6 +1090,8 @@ bool CZNC::DoRehash(CString& sError)
 		m_sStatusPrefix = sVal;
 	if (config.FindStringEntry("sslcertfile", sVal))
 		m_sSSLCertFile = sVal;
+	if (config.FindStringEntry("sslciphers", sVal))
+		m_sSSLCiphers = sVal;
 	if (config.FindStringEntry("skin", sVal))
 		SetSkinName(sVal);
 	if (config.FindStringEntry("connectdelay", sVal))
@@ -1228,6 +1104,47 @@ bool CZNC::DoRehash(CString& sError)
 		m_uiMaxBufferSize = sVal.ToUInt();
 	if (config.FindStringEntry("protectwebsessions", sVal))
   		m_bProtectWebSessions = sVal.ToBool();
+	if (config.FindStringEntry("hideversion", sVal))
+		m_bHideVersion = sVal.ToBool();
+
+	if (config.FindStringEntry("sslprotocols", m_sSSLProtocols)) {
+		VCString vsProtocols;
+		m_sSSLProtocols.Split(" ", vsProtocols, false, "", "", true, true);
+
+		for (CString& sProtocol : vsProtocols) {
+
+			unsigned int uFlag = 0;
+			bool bEnable = sProtocol.TrimPrefix("+");
+			bool bDisable = sProtocol.TrimPrefix("-");
+
+			if (sProtocol.Equals("All")) {
+				uFlag = ~0;
+			} else if (sProtocol.Equals("SSLv2")) {
+				uFlag = Csock::EDP_SSLv2;
+			} else if (sProtocol.Equals("SSLv3")) {
+				uFlag = Csock::EDP_SSLv3;
+			} else if (sProtocol.Equals("TLSv1")) {
+				uFlag = Csock::EDP_TLSv1;
+			} else if (sProtocol.Equals("TLSv1.1")) {
+				uFlag = Csock::EDP_TLSv1_1;
+			} else if (sProtocol.Equals("TLSv1.2")) {
+				uFlag = Csock::EDP_TLSv1_2;
+			} else {
+				CUtils::PrintError("Invalid SSLProtocols value [" + sProtocol + "]");
+				CUtils::PrintError("The syntax is [SSLProtocols = [+|-]<protocol> ...]");
+				CUtils::PrintError("Available protocols are [SSLv2, SSLv3, TLSv1, TLSv1.1, TLSv1.2]");
+				return false;
+			}
+
+			if (bEnable) {
+				m_uDisabledSSLProtocols &= ~uFlag;
+			} else if (bDisable) {
+				m_uDisabledSSLProtocols |= uFlag;
+			} else {
+				m_uDisabledSSLProtocols = ~uFlag;
+			}
+		}
+	}
 
 	// This has to be after SSLCertFile is handled since it uses that value
 	const char *szListenerEntries[] = {
@@ -1427,6 +1344,37 @@ bool CZNC::RemBindHost(const CString& sHost) {
 	return false;
 }
 
+void CZNC::ClearTrustedProxies() {
+	m_vsTrustedProxies.clear();
+}
+
+bool CZNC::AddTrustedProxy(const CString& sHost) {
+	if (sHost.empty()) {
+		return false;
+	}
+
+	for (unsigned int a = 0; a < m_vsTrustedProxies.size(); a++) {
+		if (m_vsTrustedProxies[a].Equals(sHost)) {
+			return false;
+		}
+	}
+
+	m_vsTrustedProxies.push_back(sHost);
+	return true;
+}
+
+bool CZNC::RemTrustedProxy(const CString& sHost) {
+	VCString::iterator it;
+	for (it = m_vsTrustedProxies.begin(); it != m_vsTrustedProxies.end(); ++it) {
+		if (sHost.Equals(*it)) {
+			m_vsTrustedProxies.erase(it);
+			return true;
+		}
+	}
+
+	return false;
+}
+
 void CZNC::Broadcast(const CString& sMessage, bool bAdminOnly,
 		CUser* pSkipUser, CClient *pSkipClient) {
 	for (map<CString,CUser*>::iterator a = m_msUsers.begin(); a != m_msUsers.end(); ++a) {
@@ -1643,12 +1591,15 @@ bool CZNC::AddListener(const CString& sLine, CString& sError) {
 		bSSL = true;
 	}
 
+	// No support for URIPrefix for old-style configs.
+	CString sURIPrefix;
 	unsigned short uPort = sPort.ToUShort();
-	return AddListener(uPort, sBindHost, bSSL, eAddr, eAccept, sError);
+	return AddListener(uPort, sBindHost, sURIPrefix, bSSL, eAddr, eAccept, sError);
 }
 
-bool CZNC::AddListener(unsigned short uPort, const CString& sBindHost, bool bSSL,
-			EAddrType eAddr, CListener::EAcceptType eAccept, CString& sError) {
+bool CZNC::AddListener(unsigned short uPort, const CString& sBindHost,
+		       const CString& sURIPrefixRaw, bool bSSL,
+		       EAddrType eAddr, CListener::EAcceptType eAccept, CString& sError) {
 	CString sHostComment;
 
 	if (!sBindHost.empty()) {
@@ -1709,7 +1660,18 @@ bool CZNC::AddListener(unsigned short uPort, const CString& sBindHost, bool bSSL
 		return false;
 	}
 
-	CListener* pListener = new CListener(uPort, sBindHost, bSSL, eAddr, eAccept);
+	// URIPrefix must start with a slash and end without one.
+	CString sURIPrefix = CString(sURIPrefixRaw);
+	if(!sURIPrefix.empty()) {
+		if (!sURIPrefix.StartsWith("/")) {
+			sURIPrefix = "/" + sURIPrefix;
+		}
+		if (sURIPrefix.EndsWith("/")) {
+			sURIPrefix.TrimRight("/");
+		}
+	}
+
+	CListener* pListener = new CListener(uPort, sBindHost, sURIPrefix, bSSL, eAddr, eAccept);
 
 	if (!pListener->Listen()) {
 		sError = FormatBindError();
@@ -1726,6 +1688,7 @@ bool CZNC::AddListener(unsigned short uPort, const CString& sBindHost, bool bSSL
 
 bool CZNC::AddListener(CConfig* pConfig, CString& sError) {
 	CString sBindHost;
+	CString sURIPrefix;
 	bool bSSL;
 	bool b4;
 #ifdef HAVE_IPV6
@@ -1747,6 +1710,7 @@ bool CZNC::AddListener(CConfig* pConfig, CString& sError) {
 	pConfig->FindBoolEntry("ipv6", b6, b6);
 	pConfig->FindBoolEntry("allowirc", bIRC, true);
 	pConfig->FindBoolEntry("allowweb", bWeb, true);
+	pConfig->FindStringEntry("uriprefix", sURIPrefix);
 
 	EAddrType eAddr;
 	if (b4 && b6) {
@@ -1774,7 +1738,7 @@ bool CZNC::AddListener(CConfig* pConfig, CString& sError) {
 		return false;
 	}
 
-	return AddListener(uPort, sBindHost, bSSL, eAddr, eAccept, sError);
+	return AddListener(uPort, sBindHost, sURIPrefix, bSSL, eAddr, eAccept, sError);
 }
 
 bool CZNC::AddListener(CListener* pListener) {
@@ -1805,9 +1769,22 @@ bool CZNC::DelListener(CListener* pListener) {
 	return false;
 }
 
+static CZNC* s_pZNC = NULL;
+
+void CZNC::CreateInstance() {
+	if (s_pZNC)
+		abort();
+
+	s_pZNC = new CZNC();
+}
+
 CZNC& CZNC::Get() {
-	static CZNC* pZNC = new CZNC;
-	return *pZNC;
+	return *s_pZNC;
+}
+
+void CZNC::DestroyInstance() {
+	delete s_pZNC;
+	s_pZNC = NULL;
 }
 
 CZNC::TrafficStatsMap CZNC::GetTrafficStats(TrafficStatsPair &Users,
@@ -1852,7 +1829,7 @@ CZNC::TrafficStatsMap CZNC::GetTrafficStats(TrafficStatsPair &Users,
 	return ret;
 }
 
-void CZNC::AuthUser(CSmartPtr<CAuthBase> AuthClass) {
+void CZNC::AuthUser(std::shared_ptr<CAuthBase> AuthClass) {
 	// TODO unless the auth module calls it, CUser::IsHostAllowed() is not honoured
 	bool bReturn = false;
 	GLOBALMODULECALL(OnLoginAttempt(AuthClass), &bReturn);
@@ -1932,6 +1909,10 @@ protected:
 };
 
 void CZNC::SetConnectDelay(unsigned int i) {
+	if (i < 1) {
+		// Don't hammer server with our failed connects
+		i = 1;
+	}
 	if (m_uiConnectDelay != i && m_pConnectQueueTimer != NULL) {
 		m_pConnectQueueTimer->Start(i);
 	}
@@ -1980,7 +1961,6 @@ void CZNC::AddNetworkToQueue(CIRCNetwork *pNetwork) {
 		}
 	}
 
-
 	m_lpConnectQueue.push_back(pNetwork);
 	EnableConnectQueue();
 }
